class Handler():
    '''Handling sqlite files from gridded obs
    '''

    def connect(self, must_exist=False, none_for_nonexistent=False, create=False):
        """connect to db with different behavior if file does not already exist

        args:
        must_exist:           if True, an error is thrown if file does not exist
        none_for_nonexistent: if True, return None        if file does not exist
        create:               if True, file is created
        """
        import os
        import sqlite3
        import warnings

        if not os.path.isfile(self.sqlite_file):
            if must_exist:
                raise RuntimeError('File: '+self.sqlite_file+ '  does not exists')

            #return None if file does not exists
            if none_for_nonexistent:
                warnings.warn('File: '+self.sqlite_file+ '  does not exists; returning None')
                return None

            if create:
                #do nothing, file is created automatically
                pass

        try:
            conn = sqlite3.connect(self.sqlite_file, timeout=30.0)
        except:
            raise RuntimeError('Problem connecting to sqlite file:'+self.sqlite_file)

        return conn

    def make_filename(self, params, date, reference_name, verified_name):
        """return sql filename for two experiments being compared
        """

        import os

        #replace '%verified_name'  with the verified  experiment name
        #replace '%reference_name' with the reference experiment name
        outname = params.outname_file_struc.replace('%verified_name', verified_name).replace('%reference_name', reference_name)

        #put date in output name
        outname = date.strftime(outname)

        return os.path.join(params.score_dir, outname)


    
    def get_hist_bin_edges(self,c):
        """get edges of histograms in file
        """
        import numpy as np
        cmd = '''select edge from hist_bin_edges ;'''
        c.execute(cmd)
        entries = c.fetchall()
        edges = np.full(len(entries), np.nan)
        for ii, entry in enumerate(entries):
            edges[ii] = entry[0]
        return edges
    
    def get_domain_id(self, c,domain_name):
        """get domain id
        """
        import warnings
    
        cmd = 'select id from domains where (dname = \''+domain_name+'\');'
        try:
            c.execute(cmd)
            domain_id = c.fetchall()
            if len(domain_id) == 0:
                warnings.warn('problem getting domain_id in '+self.sqlite_file+ 'returning None')
                return None
            else:
                if len(domain_id) > 1:
                    warnings.warn('more than one value returned for domain_id in '+self.sqlite_file+ 'returning None')
                    return None

                #this is the domain id that we want
                return domain_id[0][0]
    
        except:
            warnings.warn('problem getting domain_id in '+self.sqlite_file+ 'returning None')
            return None
    
    def get_threshold_id(self, c, threshold):
        """get threshold id
        """

        import warnings

        small = 1e-6
    
        low  = '{:10.6f}'.format(threshold - small)
        high = '{:10.6f}'.format(threshold + small)
        cmd = 'select id from thresholds where (value > '+low+' and value < '+high+');'
        try:
            c.execute(cmd)
            threshold_id = c.fetchall()
            if len(threshold_id) == 0:
                warnings.warn('problem getting threshold_id in '+self.sqlite_file+ 'returning None')
                return None
            else:
                if len(threshold_id) > 1:
                    warnings.warn('more than one value returned for threshold_id in '+self.sqlite_file+ 'returning None')
                    return None

                #this is the threshold id we are looking for
                return threshold_id[0][0]
    
        except:
            warnings.warn('problem getting threshold_id in '+self.sqlite_file+ 'returning None')
            return None

    def hist_strings_from_edges(self, hist_bin_edges, str_type=None):
        """make strings for histogram columns

        allows to create the columns, update the data and read it with select
        """

        if str_type is None:
            raise ValueError('need to define a type of string desired')
        elif str_type == 'create':
            suffix = ' real, '
        elif str_type == 'update':
            suffix = '=? , '
        elif str_type == 'select':
            suffix = ', '
        else:
            raise ValueError('unknown str_type')

        #make the string
        out_str = ''
        for ii, emin in enumerate(hist_bin_edges[:-1]):
            out_str += ('h_{:05d}'+suffix).format(ii)

        #the -2 here removes the last space and comma
        return out_str[:-2]

    def exec_wait(self, c, cmd, arg=None):
        '''exec commands in a loop to mitigate "database is locked" errors
        TODO: This is probably not the best way to handle this, 
        '''
        import time
        import random
        loop = True
        while loop:
            try:
                c.execute(cmd, arg)
                loop = False
            except:
                print('wait a little!!')
                time.sleep( random.uniform(0., 1.) )


    def get_element_id(self, c, table_name=None, domain_id=None, leadtime_minutes=None, threshold_id=None):
        '''check if the line we want to fill already exists in a given sql table
        '''


        if table_name is None:
            raise ValueError('table name is mandatory')

        if domain_id is None:
            raise ValueError('domain_id is mandatory')

        if leadtime_minutes is None:
            raise ValueError('leadtime_minutes is mandatory')

        if threshold_id is None:
            cmd = '''select id from '''+table_name+''' where (leadtime_minutes = '''+str(leadtime_minutes)+'''
                                                              and domain_id    = '''+str(domain_id)+');'
        else:
            cmd = '''select id from '''+table_name+''' where (leadtime_minutes = '''+str(leadtime_minutes)+'''
                                                              and domain_id    = '''+str(domain_id)+'''
                                                              and threshold_id = '''+str(threshold_id)+');'
        c.execute(cmd)
        d_id = c.fetchall()

        if len(d_id) == 0:
            #line does not exist, create it and keep track of id
            if threshold_id is None:
                cmd = '''insert into '''+table_name+''' (leadtime_minutes, domain_id) VALUES(?,?) ;'''
                self.exec_wait(c, cmd, (leadtime_minutes, domain_id))
            else:
                cmd = '''insert into '''+table_name+''' (leadtime_minutes, domain_id, threshold_id) VALUES(?,?,?) ;'''
                self.exec_wait(c, cmd, (leadtime_minutes, domain_id, threshold_id))
            d_id = c.lastrowid
            already_existed = False
        else:
            if len(d_id) > 1:
                raise ValueError('more than one value returned for id in histogram table ')
            d_id = d_id[0][0]
            already_existed = True

        return d_id, already_existed
    


    
    def insert_hist(self, c, table_name, element_id, histogram ,params):
        """insert histogram counts in a given histogram table
        """
    
        #insert data in table
        h_col_update_str = self.hist_strings_from_edges(params.hist_bin_edges, str_type='update')
        cmd = '''update '''+table_name+''' set '''+h_col_update_str+'''  where (id = '''+str(element_id)+'''); '''
        h_tuple = tuple([float(val) for val in histogram])  #conversion to float to make sqlite3 happy
        c.execute(cmd,h_tuple)

    def get_hist_values(self, c, table_name, leadtime_minutes, domain_id, hist_bin_edges, hist_string):
        """read histogram counts in a given histogram table
        """
        import numpy as np
    
        #id of element we need
        element_id, is_there = self.get_element_id(c, table_name=table_name, domain_id=domain_id, leadtime_minutes=leadtime_minutes)
        if is_there:
            #read data in table
            cmd = '''select '''+hist_string+''' from '''+table_name+'''  where (id = '''+str(element_id)+'''); '''
            c.execute(cmd)
            entries = np.array(c.fetchall()[0])
            return entries 
        else:
            #no histogram at desired leadtime
            return None

    def populate_id_dict(self, c, params):
        '''return dictionary of domains:domain_id and threshold:threshold_id
        '''

        #for domains
        domain_id_dict = {}
        for domain in params.verif_domains:
            #id element for the domain being veridied
            domain_id = params.sql_handler.get_domain_id(c,domain)
            if domain_id is None:
                raise ValueError('Problem populatinf domain_ids dictionary')
            domain_id_dict[domain] = domain_id
        self.domain_id_dict = domain_id_dict
    

        threshold_id_dict = {}
        for threshold in params.thresholds:
            #id element for the threshold being veridied
            threshold_id = params.sql_handler.get_threshold_id(c,threshold)
            if threshold_id is None:
                raise ValueError('Problem populatinf domain_ids dictionary')
            threshold_id_dict[str(threshold)] = threshold_id
        self.threshold_id_dict = threshold_id_dict
    
    
    def init_sqlite_file(self, params):

        '''#initialize sql tables that will contain the data
        '''

        import sqlite3
        import os
        import domutils._py_tools as py_tools

        if os.path.isfile(self.sqlite_file):
            if params.complete_mode == 'complete':
                #if file exists and mode is continue
                #make sure all domains and thresholds are in list

                conn = params.sql_handler.connect(must_exist=True)
                c = conn.cursor()

                #lets make sure that all domains are in domains table
                for domain in params.verif_domains:
                    #id element for the domain being veridied
                    domain_id = params.sql_handler.get_domain_id(c,domain)
                    if domain_id is None:
                        cmd = '''insert into domains (dname) VALUES (?);'''
                        c.execute(cmd, (domain,))

                #lets make sure that all thresholds are in thresholds table
                for threshold in params.thresholds:
                    #id element for the thresholds being veridied
                    threshold_id = params.sql_handler.get_threshold_id(c,threshold)
                    if threshold_id is None:
                        cmd = '''insert into thresholds (value) VALUES (?)'''
                        c.execute(cmd, (threshold,))

                #fill attribures containing domain and threshold dicts
                self.populate_id_dict(c, params)

                conn.commit()
                conn.close()

                return




            elif params.complete_mode == 'clobber':
                #if file exists and mode is 'clobber' erase existing file and create a new empty one
                os.remove(self.sqlite_file)
            else:
                raise ValueError('complete mode can only be "complete" or "clobber".')

        #make dir if it does not already exists
        py_tools.parallel_mkdir(os.path.dirname(self.sqlite_file))

        conn = self.connect(create=True) 
        c = conn.cursor()
        
        #
        #make info table
        cmd = '''create table if not exists info(id                integer primary key, 
                                                 reference_name    text,
                                                 verified_name     text,
                                                 min_qi            real,
                                                 grid_dx           real );'''
        c.execute(cmd)
        cmd = '''insert into info (reference_name,
                                   verified_name ,
                                   min_qi,
                                   grid_dx ) VALUES (?,?,?,?);'''
        c.execute(cmd, (params.reference_reader.name, 
                        params.verified_reader.name, 
                        params.min_qi, 
                        params.grid_dx))
        
        #
        #make threshold table
        cmd = '''create table if not exists thresholds(id    integer primary key, 
                                                       value real);'''
        c.execute(cmd)
        cmd = '''insert into thresholds (value) VALUES (?)'''
        for threshold in params.thresholds:
            c.execute(cmd, (threshold,))
        
        #
        #make domains table
        cmd = '''create table if not exists domains(id    integer primary key, 
                                                    dname text);'''
        c.execute(cmd)
        cmd = '''insert into domains (dname) VALUES (?);'''
        for domain in params.verif_domains:
            c.execute(cmd, (domain,))
        
        #
        #make histograms tables
        #
        # first a table for bin edges values
        cmd = '''create table if not exists hist_bin_edges(id  integer primary key, 
                                                           edge real );'''
        c.execute(cmd)
        cmd = '''insert into hist_bin_edges (edge) VALUES (?);'''
        for edge in params.hist_bin_edges:
            c.execute(cmd, (edge,))

        # second a table for histograms from the reference dataset
        h_col_create_str = self.hist_strings_from_edges(params.hist_bin_edges, str_type='create')
        #table for reference histograms
        cmd = '''create table if not exists hist_reference(id  integer primary key, 
                                                           leadtime_minutes integer, 
                                                           domain_id        integer,
                                                           '''+h_col_create_str+');'
        c.execute(cmd)
        #
        # third a table for histograms from the verified dataset
        cmd = '''create table if not exists hist_verified( id  integer primary key, 
                                                           leadtime_minutes integer, 
                                                           domain_id        integer,
                                                           '''+h_col_create_str+');'
        c.execute(cmd)
        
        
        
        #
        #make stats table to be filled by processes
        cmd = '''create table if not exists stats(id  integer primary key, 
                                                  leadtime_minutes integer, 
                                                  threshold_id     integer,
                                                  domain_id        integer,
                                                  x                integer,
                                                  y                integer,
                                                  z                integer,
                                                  w                integer,
                                                  lmin             real,
                                                  corr_coeff       real );'''
        c.execute(cmd)

        
        #fill attribures containing domain and threshold dicts
        self.populate_id_dict(c, params)

        conn.commit()
        conn.close()



    def __init__(self, date, params, reference_name, verified_name):

        '''#initialize object that will handle sqlite files
        '''

        sqlite_file = self.make_filename(params, date, 
                                         reference_name, 
                                         verified_name)


        #
        #
        #object attributes for later use
        self.sqlite_file = sqlite_file




def reader_init(reader_type, available_args):
    """initialise reader object

    Readers usually need some information not shared by other readers
    For example, some readers search many files for a valid entries while others
    will need instruction to build specific filenames.

    Moreover, the same reader can be used irrespective of the fact that a quantity
    is considered a 'reference' or the 'verified' quantity.
    Consider how radar data can be a reference when verifying model forecast while it may also be verified 
    against other precipitation products. 

    In the call to gridded_obs, by convention:
        - all variables specific to the 'reference' reader will have the 'reference_' prefix
        - all variables specific to the 'verified'  reader will have the 'verified_'  prefix

    In this function, a reader object is created by:
        - make a dictionary containing all the 'reference' or 'verified' arguments from args object
        - the 'reference_' or 'verified_' prefix are removed so that readers can be used in either cases

    args:
        reader_type:    [str] either 'reference' or 'verified'
        available_args: [object] the "unknown_args" produced by parsing arguments in gridded_obs
                                 contains both the 'reference_' and 'verified_' arguments

    returns:
        A reader object 
    
    """

    import gridded_obs.readers as readers

    if not ((reader_type == 'reference') or (reader_type == 'verified')):
        raise ValueError('reader_type may only be only "reference" or "verified"')

    #add underscore to reader type
    reader_type += '_'

    #make list of arguments for reference or verified function 
    #prefix is removed, and variables not related to 'reader_type' are discarted
    #
    #for example with reader_type='verified' ; 
    #   ['--verified_pr_dt', '10', '--reference_data_dir', '/path/to/data/dir']
    #becomes
    #   {'pr_dt':'10'}

    #the following assumes only one key-value pair 
    #situation such as ['--verified_argument', '10', '20', '30'] will not be handled correctly
    arg_dict = {}
    for ii, element in enumerate(available_args):
        if element.startswith('--'+reader_type):
            #element is an argument of desired type
            #save argument with prefix removed
            key = element[len(reader_type)+2:]
            value = available_args[ii+1]
            arg_dict[key] = value

    #get reader function
    this_reader = getattr(readers, arg_dict['reader'])

    #remove reader from argument dict
    del arg_dict['reader']

    return this_reader(**arg_dict)











#implementation of parallel FSS for Lmin in HRDPS in "resonable" time


    #np.set_printoptions(threshold=np.inf)

import dask
@dask.delayed
def dask_partial_fractions(*args, **kwargs):
    return partial_fractions(*args, **kwargs)

def partial_fractions(radius_km, 
                      this_batch, latitudes, longitudes,
                      kdtree, reference_true, reference_qi, verified_true, 
                      small=1e-3):

    """compute FSS fractions for reference and verified fields

    uses kdtree to find all neighbors within a given radius
    arguments:
        radius_km:      [km] radius of circle for computing FSS
        this_batch:     [1D ndarray] indices in lat/lon arrays to process 
        latitudes:      [1D ndarray] latitudes of grid for estimating fractions
        longitudes:     [1D ndarray] longitudes of grid for estimating fractions
        kdtree:         [object] kdtree object for finding data points with observations within a certain radius
        reference_true: [1D ndarray] observation precip rate > threshold    ;length of this array matches lat/lon used in building kdtree
        reference_qi:   [1D ndarray] observation quality index [0,1]        ;length of this array matches lat/lon used in building kdtree
        verified_true:  [1D ndarray] verified precip rate > threshold       ;length of this array matches lat/lon used in building kdtree
        small:          [float] values below which sum of quality index is considered zero
    """

    import time
    import numpy as np

    #convert lat/lon of points to search to xyz
    ii_s = this_batch[0,:]
    jj_s = this_batch[1,:]
    batch_xyz = to_xyz(longitudes[ii_s,jj_s], latitudes[ii_s,jj_s])

    #initialize output array
    reference_fractions = np.zeros_like(latitudes)
    verified_fractions  = np.zeros_like(latitudes)
    
    #list of neighboring points
    radius_m = 1e3*radius_km
    for ii, jj, xyz in zip(ii_s, jj_s, batch_xyz):
        #for some reason, calling kdtree outside the loop makes things slower...
        ball_list = kdtree.query_ball_point(xyz, radius_m)

        #record fractions
        if len(ball_list) == 0:
            #no valid datapoints
            reference_fractions[ii,jj] = -1.
            verified_fractions[ii,jj]  = -1.
        else:
            #generate fractions
            w_sum = np.sum(reference_qi[ball_list])
            #only normalize if sum of weights is larger than a small value. Otherwise, no data of sufficient quality was present
            if w_sum > small:
                reference_fractions[ii,jj] = np.sum(reference_qi[ball_list]*reference_true[ball_list]) / w_sum
                verified_fractions[ii,jj]  = np.sum(reference_qi[ball_list]* verified_true[ball_list]) / w_sum
            else:
                reference_fractions[ii,jj] = -1.
                verified_fractions[ii,jj]  = -1.

    return np.dstack((reference_fractions, verified_fractions))


def fss_diff(radius, target_fss, mode, 
             reference_true, reference_qi, verified_true, grid_dx,
             n_cpus=1, verif_longitudes=None, verif_latitudes=None,
             small = 1e-3, debug=False):

    """difference between FSS as a given radius and pre-computed FSS target

    This function is to be used by a minimizer to find smallest lenghtscale at which 
    a forecast is usefull

    Three "modes" or methods for computing FSS can be used
    - fft               by far the fastest; assumes regularly gridded data with a quality index
    - manual_parallel   Manual convolution performed in parallel; accepts irregularly gridded data defined by lat/lon
    - manual_serial     Manual convulution performed serially

    The three methods were verified to give the same FSS +- 0.01

    args:
    radius:                 [float km] radius of circle for computing FSS
    target_fss:             [float]    FSS value above which a forecast is deemed "skilled"
    mode:                   [str]      method to use for FSS computation; see above
    reference_true:         [2d ndarray] 1 where references values > threshold, 0 otherwise
    reference_qi:           [2d ndarray] quality index of reference data
    verified_true:          [2d ndarray] 1 where rverified values > threshold, 0 otherwise
    grid_dx:                [float km] size of grid tiles, assumed constant
    n_cpus:                 [int] only for method "manual_parallela; " number of cpus 
    verif_longitudes:       [ndarray] longitudes of verification grid
    verif_latitudes:        [ndarray] latitudes  of verification grid
    small:                  [float] value below which sum of quality index if considered zero
    debug:                  [bool] output fractions for debugging

    """

    import numpy as np
    import time
    import dask
    import dask.array as da
    import scipy.signal

    if not ((mode == 'fft') or (mode == 'manual_serial') or (mode == 'manual_parallel')) :
        raise ValueError('Unrecognized mode')

    if mode == 'fft':
        #weighted fractions through fft convolution

        #other possible kernels
        #   kernel = np.exp(-1.*np.sqrt(xx**2.+yy**2.)/.02)
        #   kernel = np.outer(scipy.signal.windows.gaussian(npts, 3), scipy.signal.windows.gaussian(npts, 3))
        #
        #making uniform circular kernel
        npts = np.int(2.*(radius/grid_dx)+1.)
        xx, yy = np.meshgrid(np.linspace(-1.,1.,npts),np.linspace(-1.,1.,num=npts))
        kernel = np.where(np.sqrt(xx**2. + yy**2.) <= 1., 1., 0.)

        #convolve quality index
        qi_convol        = scipy.signal.fftconvolve(reference_qi,    kernel, mode='same')

        #qi_convol will be used as a denominator
        #these indices mark values that large enough to be used
        #also get rid of -ve values that occur due to noise in the convolution
        inds_f = (qi_convol > small).nonzero()

        #-ve values in output will not be used for FSS computation
        reference_fractions = np.full_like(qi_convol, -1.)  
        verified_fractions  = np.full_like(qi_convol, -1.)  

        #convolve and normalize
        #
        reference_pr_qi_convol = scipy.signal.fftconvolve(reference_true * reference_qi, kernel, mode='same') 
        reference_fractions[inds_f] = reference_pr_qi_convol[inds_f] / qi_convol[inds_f]
        #
        verified_pr_qi_convol  = scipy.signal.fftconvolve( verified_true * reference_qi, kernel, mode='same') 
        verified_fractions[inds_f]  = verified_pr_qi_convol[inds_f]  / qi_convol[inds_f]


    else:
        #sample everything where observations are valid
        #verification can only be performed in the presence of observations
        inds = np.asarray(reference_qi > 0.).nonzero()
        if inds[0].size == 0:
            warnings.warn('no observations with quality index greater than threshold in the verification domain. Returning lmin=NaN')
            return np.nan
        sampled_latitudes      =  verif_latitudes[inds]
        sampled_longitudes     = verif_longitudes[inds]
        sampled_reference_true =   reference_true[inds]
        sampled_reference_qi   =     reference_qi[inds]
        sampled_verified_true  =    verified_true[inds]
        orig_x = np.arange(0, verif_latitudes.shape[0])
        orig_y = np.arange(0, verif_latitudes.shape[1])
        orig_xx, orig_yy = np.meshgrid(orig_x,orig_y,indexing='ij')
        orig_inds = np.vstack((orig_xx.ravel(), orig_yy.ravel()))

        #build kdtree using all points that can be used in the verification
        grid_xyz = to_xyz(sampled_longitudes, sampled_latitudes)
        kdtree = scipy.spatial.cKDTree(grid_xyz, balanced_tree=False)

        #divide this list into chuncks that fit the number of processors we have for the job
        splitted_ij = np.array_split(orig_inds, n_cpus, axis=1)

        if mode == 'manual_serial':

            reference_fractions = np.zeros_like(verif_latitudes)
            verified_fractions =  np.zeros_like(verif_latitudes)
            for this_batch in splitted_ij:
                total = partial_fractions(radius, 
                                          this_batch, verif_latitudes, verif_longitudes, 
                                          kdtree, sampled_reference_true, sampled_reference_qi, sampled_verified_true)
                this_ref_frac   = total[:,:,0]
                this_verif_frac = total[:,:,1]
                reference_fractions += this_ref_frac
                verified_fractions  += this_verif_frac

        elif mode == 'manual_parallel':

            # sample output so dask knows what to expect only shape and type matters here
            sample = np.dstack((verif_latitudes, verif_longitudes))
            #list of desired outputs
            res_list = [da.from_delayed(dask_partial_fractions(radius, 
                                                               this_batch, verif_latitudes, verif_longitudes, 
                                                               kdtree, sampled_reference_true, sampled_reference_qi, sampled_verified_true), 
                                                               sample.shape, sample.dtype) for this_batch in splitted_ij]
            res_array = da.stack(res_list, axis=3)
            # expected result
            res = res_array.sum(axis=3)
            # Computation is performed here.
            total = res.compute()
            reference_fractions = total[:,:,0]
            verified_fractions  = total[:,:,1]



    #compute FSS
    #remove all -ve values from FSS computation
    inds = np.logical_and(reference_fractions >= -small, verified_fractions >= -small).nonzero()
    mse     = np.nansum( np.square(reference_fractions[inds]  - verified_fractions[inds]) )
    mse_ref = np.nansum( np.square(reference_fractions[inds]) + np.square(verified_fractions[inds]) )
    if np.isclose(mse_ref, 0.):
        #no data is present or too small a search radius
        this_fss = 0.
        fss_difference = - 1.
    else:
        this_fss = 1. - mse / mse_ref
        fss_difference = this_fss - target_fss

    #print(mode, ('{:8.4f}'*4).format(radius, this_fss, target_fss, fss_difference))

    #return difference with FSS target
    if debug:
        return fss_difference, reference_fractions, verified_fractions
    else:
        #normal mode
        return fss_difference


def to_xyz(lons, lats):
    import numpy as np
    import scipy.spatial
    import cartopy.crs as ccrs

    from domutils.geo_tools.lat_lon_extend import lat_lon_extend
    """ From lat/lon to xyz on unit sphere
    
    Args:
        lons, lats : coords to be converted to xyz
    """

    #insure numpy arrays
    xx=np.asarray(lons)
    yy=np.asarray(lats)

    #everything to xyz coords 
    proj_ll = ccrs.Geodetic()
    geo_cent = proj_ll.as_geocentric()

    return geo_cent.transform_points(proj_ll,
                                     xx.flatten(),
                                     yy.flatten())


def lmin_from_fss(reference_true, reference_qi, verified_true, mode,
         lmin_range,
         latitudes=None, longitudes=None, 
         grid_dx=None, n_cpus=1, use_qi=True, dask_client_exists=False):

    """minimum scale at which forecast has skill

    lmin from Fraction Skill Score

    it is assumed that reference an verified data are already on the same verification grid

    arguments:
        reference_true: [unitless] 1 when reference field > threshold; 0 otherwise
        reference_qi:   [unitless] quality index of observations
        verified_true:  [unitless] 1 when verified  field > threshold; 0 otherwise
        mode:           [str]   name of method to used for computing FSS, see 'fss_diff' doc for valuid options
        lmin_range:     [min,max] range of radius between which lmin will be searched
        latitudes:      [degrees] latitude of grid  ; only for "manual" methods
        longitudes:     [degrees] longitude of grid ; only for "manual" methods
        grid_dx:        [km] size of grid; tile necessary for 'fft' method
        n_cpus:         Number of cpus we have for the job only used for 'manual_parallel' method
        dask_client_exists: if True, no attempts will be made at initializing dask client
        
    """
    #import os
    #os.environ["OMP_NUM_THREADS"] = "1"        # export OMP_NUM_THREADS=1
    #os.environ["OPENBLAS_NUM_THREADS"] = "1"   # export OPENBLAS_NUM_THREADS=1
    #os.environ["MKL_NUM_THREADS"] = "1"        # export MKL_NUM_THREADS=1
    #os.environ["VECLIB_MAXIMUM_THREADS"] = "1" # export VECLIB_MAXIMUM_THREADS=1
    #os.environ["NUMEXPR_NUM_THREADS"] = "1"    # export NUMEXPR_NUM_THREADS=1   

    import sys
    import os
    import time
    import warnings
    import numpy as np
    import scipy.spatial
    import scipy.optimize
    import dask.distributed


    #make sure that all inputs have same shape
    if (reference_true.shape != reference_qi.shape) or (reference_true.shape != verified_true.shape) :
        raise ValueError('all inputs must have same shape')

    #target fss for a "usefull" forecast
    target_fss = 0.5 + ( np.sum(reference_qi*reference_true) / np.sum(reference_qi) )/2.

    #initialize dask client if needed
    if (mode == 'manual_parallel') and (not dask_client_exists):
        #parallel execution, initialize dask client
        client = dask.distributed.Client(processes=True, threads_per_worker=1, n_workers=n_cpus, silence_logs=40)


    #find lmin 
    try:
        lmin_v, r = scipy.optimize.brentq(fss_diff, lmin_range[0], lmin_range[1], 
                                        args=(target_fss, mode,
                                              reference_true, reference_qi, verified_true, grid_dx),
                                        xtol=1., rtol=1e-6, maxiter=100, full_output=True, disp=False)
        #get out of loop if we converged
        if not r.converged:
            print('lmin did not converge ')
            lmin_v = None
     
    except KeyboardInterrupt:
        print('Interrupted')
        try:
            sys.exit(1)
        except SystemExit:
            os._exit(1)
    except:
        print('lmin did not converge ')
        lmin_v = None

    ##for debugging
    #lmin_v = 1.
    #lmin_min = 51.
    #lmin_max = 1000
    #for radius in np.arange(lmin_min,lmin_max, 10):
    #    ##method 1
    #    #mode = 'manual_serial'
    #    #t1 = time.time()
    #    #this_fss_diff = fss_diff(radius, target_fss, mode, 
    #    #                         reference_true, reference_qi, verified_true, grid_dx,
    #    #                         verif_longitudes=longitudes, verif_latitudes=latitudes)
    #    #t2 = time.time()
    #    #print(mode, 'time: ', t2-t1, 'fss_diff', this_fss_diff)

    #    #method 2
    #    mode = 'manual_parallel'
    #    t1 = time.time()
    #    this_fss_diff, ref_frac_parallel, verif_frac_parallel = fss_diff(radius, target_fss, mode, 
    #                                                                     reference_true, reference_qi, verified_true, grid_dx,
    #                                                                     verif_longitudes=longitudes, verif_latitudes=latitudes, 
    #                                                                     n_cpus=40, debug=True)
    #    t2 = time.time()
    #    #print(mode, 'time: ', t2-t1, 'fss_diff', this_fss_diff)

    #    #method 3
    #    mode = 'fft'
    #    t1 = time.time()
    #    this_fss_diff, ref_frac_convol, verif_frac_convol = fss_diff(radius, target_fss, mode, 
    #                                                                 reference_true, reference_qi, verified_true, grid_dx,
    #                                                                 debug=True)
    #                             
    #    t2 = time.time()
    #    #print(mode, 'time: ', t2-t1, 'fss_diff', this_fss_diff)

    #    ##show image for verification
    #    import compare_fields
    #    compare_fields.plot_fraction(reference_true, longitudes, latitudes, 
    #                                 rf=ref_frac_parallel,    rfi=ref_frac_convol,
    #                                 vf=verif_frac_parallel,vfi=verif_frac_convol,
    #                                 radius=radius)
    #    exit()



    return lmin_v




def main():

    import datetime
    import numpy as np
    import domcmc.fst_tools as fst_tools



    ##Example 1---------------------------------------------------------------
    ##variables controlling the verification
    #threshold = .1  #[mm/h]
    #min_qi = .2
    #lmin_max=10

    ##variables controlling the execution
    #n_cpus = 3

    ##data
    #latitudes = np.array([[51.82844 , 51.850636, 51.872826, 51.895023],
    #                      [51.824722, 51.846912, 51.869102, 51.89129 ],
    #                      [51.820988, 51.84318 , 51.865364, 51.887558],
    #                      [51.817253, 51.839447, 51.861626, 51.88382 ]])
    #longitudes= np.array([[257.0859 , 257.09192, 257.09793, 257.10394],
    #                      [257.1218 , 257.1278 , 257.13385, 257.1399 ],
    #                      [257.15768, 257.16376, 257.1698 , 257.17584],
    #                      [257.19354, 257.19962, 257.2057 , 257.21176]])

    #reference_pr = np.array([[ 0,  0,  0,  0],
    #                         [ 0,  1,  0,  0],
    #                         [ 0,  0,  0,  0],
    #                         [ 0,  0,  0,  0]])

    ##reference_qi = np.array([[ 0,  0,  0,  0],
    ##                         [ 0,  1,  0,  0],
    ##                         [ 0,  0,  0,  0],
    ##                         [ 0,  0,  0,  0]])

    ##reference_qi = np.array([[ 0,  0,  0,  0],
    ##                         [ 0,  1,  1,  0],
    ##                         [ 0,  1,  1,  0],
    ##                         [ 0,  0,  0,  0]])

    #reference_qi = np.array([[ 1,  1,  1,  1],
    #                         [ 1,  1,  1,  1],
    #                         [ 1,  1,  1,  1],
    #                         [ 1,  1,  1,  1]])

    #verified_pr  = np.array([[ 0,  0,  0,  0],
    #                         [ 0,  0,  0,  0],
    #                         [ 0,  0,  1,  0],
    #                         [ 0,  0,  0,  0]])
    #
    #lmin(reference_pr, reference_qi, verified_pr, 
    #     latitudes, longitudes, 
    #     threshold, min_qi, 
    #     lmin_max,
    #     grid_dx,
    #     n_cpus)

    



    #Example 2---------------------------------------------------------------
    #reference data
    this_date = datetime.datetime(2016,7,1,1,00)
    
    radar_dir = '/space/hall4/sitestore/eccc/mrd/rpndat/dja001/lhn_input/r_data_h5_v8_smooth_04_median_03_min_natgrid2.5/'
    radar_recipe = '%Y/%m/%d/%Y%m%d%H%M_mosaic.fst'
    reference_file = radar_dir+this_date.strftime(radar_recipe)
    
    reference_pr_dict = fst_tools.get_data(file_name=reference_file, var_name='RDPR', latlon=True)
    reference_qi_dict = fst_tools.get_data(file_name=reference_file, var_name='RDQI')

    #select data only on verification grid
    domain_dict = {'radars':[200,  50,   2300, 700] }
    imin,jmin,imax,jmax = domain_dict['radars']
    reference_pr = reference_pr_dict['values'][imin:imax,jmin:jmax]
    reference_qi = reference_qi_dict['values'][imin:imax,jmin:jmax]
    latitudes  = reference_pr_dict['lat'][imin:imax,jmin:jmax]
    longitudes = reference_pr_dict['lon'][imin:imax,jmin:jmax]
    grid_dx = 2.5

    
    ##forecast data
    prefix = '2016070100' #start time of forecast
    fcst_output_dir = '/space/hall3/sitestore/eccc/mrd/rpndat/dja001/maestro_archives/N2CC801E16V1/gridpt/prog/lhn/'
    #get precipitation rate from PR diff
    fcst_pr_t_dict   = fst_tools.get_data(dir_name=fcst_output_dir, var_name='PR', prefix=prefix, datev=this_date)
    deltat = 10.*60. #[seconds] -> 10 min 
    date_mdt = this_date - datetime.timedelta(seconds=deltat)
    fcst_pr_mdt_dict = fst_tools.get_data(dir_name=fcst_output_dir, var_name='PR', prefix=prefix, datev=date_mdt)
    
    #forecast precip rate in mm/h
    #       *1000   -> conversion from meters to mm
    #       *3600   -> for precip rate during one hour
    #       /deltat -> time difference between the two accumulation (PR) values that were read
    verified_pr = (fcst_pr_t_dict['values'] - fcst_pr_mdt_dict['values'])*1000.*3600/deltat
    verified_pr = verified_pr[imin:imax,jmin:jmax]
    
    #variables controlling the verification
    threshold = .1  #[mm/h]
    min_qi = .2
    #use None for inclusion of end point
    lmin_range=[grid_dx, 500]

    #quality index < min_qi set to zero
    reference_qi   = np.where(reference_qi < min_qi,  0., reference_qi)

    #apply threshold on quantities
    reference_true = np.where(reference_pr >= threshold, 1., 0.)
    verified_true  = np.where(verified_pr  >= threshold, 1., 0.)

    ##variables controlling the execution
    n_cpus = 40

    #import dask.distributed
    #client = dask.distributed.Client(processes=True, threads_per_worker=1, n_workers=n_cpus, silence_logs=40)


    lminv = lmin(reference_true, reference_qi, verified_true, 'fft',
                 lmin_range, grid_dx=grid_dx)
                 #latitudes=latitudes, longitudes=longitudes,
                 #n_cpus=n_cpus, dask_client_exists=True)

    print('converged to lmin = ', lminv)

if __name__ =='__main__':
    main()



class ModelPrDiff:
    
    def __init__(self, name=None, data_dir=None, pr_dt=None, prefix=None):
        """Estimate pecipitation rates by substracting PR at two different times
    
        args:
            data_dir:  [str]  /basse/dir/where/data/is/
            pr_dt:     [minutes] time for PR at previous time
            prefixt:   [strftime format] restrict search to files with this format
        """

        print('initializing "modelPrDiff" reader')

        if name is not None:
            self.name  = name
        else:
            raise ValueError('keyword "name" must be speficied')

        if data_dir is not None:
            self.data_dir  = data_dir
        else:
            raise ValueError('keyword "data_dir" must be speficied')

        if pr_dt is not None:
            self.pr_dt  = pr_dt
        else:
            raise ValueError('keyword "pr_dt" must be speficied')

        if prefix is not None:
            self.prefix  = prefix
        else:
            raise ValueError('keyword "prefix" must be speficied')



    def get_data(self, validity_date, leadtime=None):
        """read data from a standard file by substracting PR at two times

        args:
            validity_date:  [datetime object]  date at which data is desired
            leadtime:       [datetime timedelta object]  offset with respect to validity time

        """
        import datetime
        import numpy as np
        import domcmc.fst_tools 

        #take leadtime into account
        if leadtime is None:
            this_date = validity_date
        else:
            this_date = validity_date + leadtime
    
        #beginning of filename
        this_prefix = validity_date.strftime(self.prefix)

        #get precipitation accumulation at validity date
        pr_t_dict   = domcmc.fst_tools.get_data(dir_name=self.data_dir, var_name='PR', prefix=this_prefix, datev=this_date, latlon=True)

        #get precipitation accumulation at validity date - pr_dt
        deltat = np.float(self.pr_dt)*60. #convert minutes to seconds
        date_mdt = this_date - datetime.timedelta(seconds=deltat)
        pr_mdt_dict = domcmc.fst_tools.get_data(dir_name=self.data_dir, var_name='PR', prefix=this_prefix, datev=date_mdt)
    
        if (pr_t_dict is None) or (pr_mdt_dict is None):
            return None
        else:
            #forecast precip rate in mm/h
            #       *1000   -> conversion from meters to mm
            #       *3600   -> for precip rate during one hour
            #       /deltat -> time difference between the two accumulation (PR) values that were read
            forecast_pr = (pr_t_dict['values'] - pr_mdt_dict['values'])*1000.*3600/deltat

            return {'values':forecast_pr,
                    'qi_values':None,
                    'lats':pr_t_dict['lat'], 
                    'lons':pr_t_dict['lon']}




class ModelFst:
    def __init__(self, name=None, data_dir=None, file_struc=None, varname=None, qi_varname=None):
        """reader for CMC 'standard' files

        args:
            data_dir:      [str]  /basse/dir/where/data/is/
            file_struc:    [strftime format]  how to construct filename from date
            varname:       [str] Name of variable to read in standard file
        """

        print('initializing "ModelFst" reader')

        if name is not None:
            self.name  = name
        else:
            raise ValueError('keyword "name" must be speficied')

        if data_dir is not None:
            self.data_dir  = data_dir
        else:
            raise ValueError('keyword "data_dir" must be speficied')

        if file_struc is not None:
            self.file_struc  = file_struc
        else:
            raise ValueError('keyword "file_struc" must be speficied')

        if varname is not None:
            self.varname  = varname
        else:
            raise ValueError('keyword "varname" must be speficied')

        #optionnal variable
        self.qi_varname = qi_varname

    def get_data(self, validity_date, leadtime=None):
        """read data from a standard file

        args:
            validity_date:  [datetime object]  date at which data is desired
            leadtime:       [datetime timedelta object]  offset with respect to validity time

        """

        import domcmc.fst_tools 
        import numpy as np

        #take leadtime into account
        if leadtime is None:
            this_date = validity_date
        else:
            this_date = validity_date + leadtime

        this_file = self.data_dir + this_date.strftime(self.file_struc)

        data_dict = domcmc.fst_tools.get_data(file_name=this_file, var_name=self.varname, datev=this_date, latlon=True)

        #get quality index 
        if self.qi_varname is not None:
            qi_dict = domcmc.fst_tools.get_data(file_name=this_file, var_name=self.qi_varname, datev=this_date)
            if qi_dict is not None:
                qi_values = qi_dict['values']
            else:
                qi_values = None
        else:
            qi_values = None


        if data_dict is None:
            return None
        else:
            return {'values':data_dict['values'], 
                    'qi_values':qi_values,
                    'lats':data_dict['lat'], 
                    'lons':data_dict['lon']}





"""functions common to verification and aggregation
"""

def make_date_list(params):
    """list of dates over which to iterate
    """

    import datetime
    import numpy as np

    if ((params.date_f == None) or
        (params.date_0 == params.date_f)):
        date_list = [params.date_0]
    else:
        t_diff = (params.date_f-params.date_0) + datetime.timedelta(seconds=1)    #+ 1 second for inclusive end point
        elasped_seconds = t_diff.days*3600.*24. + t_diff.seconds                  #I hate datetime objects...
        delta_date_seconds = params.delta_date*60.
        date_list = [params.date_0 + datetime.timedelta(seconds=x) for x in np.arange(0,elasped_seconds,delta_date_seconds)]

    return date_list

def make_leadtime_list(params):
    """list of leadtimes (datetime.timedelta) that will be verified in parallel
    """

    import datetime
    import numpy as np

    if params.leadtime_0 == params.leadtime_f:
        leadtime_list = [datetime.timedelta(seconds=params.leadtime_0*60.)]
    else:
        leadtime_list = [datetime.timedelta(seconds=lt*60.) for lt in np.arange(params.leadtime_0,params.leadtime_f, params.delta_leadtime)]

    return leadtime_list

def str2bool(v):
    import argparse
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')




"""ploting routines
"""

def plot_header( fig, domain, params, 
                 fig_w  , fig_h  ,
                 rec_w  , rec_h  ,
                 sp_w   , sp_h   ,
                 pal_sp , pal_w  ,
                 tit_h  , 
                 threshold=None):
    '''header for figure
    '''

    tit_w = 8./fig_w
    pos = [0.,1.-tit_h,tit_w,tit_h] 
    y_top = 0.95
    linespace = .5/(tit_h*fig_h)
    tit_ax = fig.add_axes(pos, zorder=10)
    tit_ax.set_facecolor((0.,0.,0.,0.))
    tit_ax.set_xlim((0.,1.))
    tit_ax.set_ylim((0.,1.))
    tit_ax.axis('off')
    #date range
    date_str = params.date_list[0].strftime('%Y%m%d%H')+' - '+params.date_list[-1].strftime('%Y%m%d%H')
    tit_ax.annotate(date_str,(sp_w, y_top-0.*linespace),           xycoords='axes fraction', ha='left', va='top', fontsize=24)
    #number of forecasts
    num_str = '{:5d}'.format(len(params.date_list))+ ' forecasts considered'
    tit_ax.annotate(num_str, (.9, y_top-0.*linespace), xycoords='axes fraction', ha='left', va='top', fontsize=24)
    #thresholds
    if threshold is not None:
        th_str = 'threshold: '+'{:10.4f}'.format(threshold)
        tit_ax.annotate(th_str, (sp_w, y_top-1.*linespace), xycoords='axes fraction', ha='left', va='top', fontsize=24)
    #domain
    do_str = 'on domain: '+domain
    tit_ax.annotate(do_str, (.9, y_top-1.*linespace), xycoords='axes fraction', ha='left', va='top', fontsize=24)
    #legend for identifying experiments
    l_top = y_top - 1.5/(tit_h*fig_h)
    xs = [.1, .8 ]
    ys = [l_top, l_top-.5/(tit_h*fig_h), l_top-1./(tit_h*fig_h)]
    pos_list = [(xx,yy) for yy in ys for xx in xs]
    line_l = .17
    for kk, experiment in enumerate(params.exp_list):
        xx = pos_list[kk][0]
        yy = pos_list[kk][1]
        tit_ax.plot(   [xx, xx+line_l], [yy, yy], 
                       color=params.exp_color[kk], 
                       linestyle=params.exp_linestyle[kk], 
                       linewidth=2.*params.exp_linewidth[kk], )
        tit_ax.annotate(experiment, (xx+line_l, yy), xycoords='data', ha='left', va='center', fontsize=24)

def autorange(score):
    '''determine nim max range for multiple experiments
    '''

    import numpy as np

    minv = np.nanmin(score)
    maxv = np.nanmax(score)
    rangev = maxv - minv
    headspace = .10 # % headspace below and above max value; improves readability

    return (minv - headspace*rangev, maxv + headspace*rangev )












import dask
@dask.delayed
def dask_verify_date_lt(*args, **kwargs):
    return verify_date_lt(*args, **kwargs)


def verify_date_lt(params, validity_date, leadtime):
    """ compute verifications scores between two gridded products
    
    """

    import os
    import datetime
    import warnings
    import gc
    import sqlite3
    from scipy import stats

    import gridded_obs.lmin_from_fss
    import numpy as np
    #import gridded_obs.plot.compare_fields

    leadtime_minutes = np.int(leadtime.days*1440. + leadtime.seconds/60.)


    #read reference field
    reference_dict = params.reference_reader.get_data(validity_date, leadtime)
    if (reference_dict is None):
        warnings.warn('Reference dataset unavailable for: ' + str(validity_date) + ' + ' + str(leadtime))

    #read quantity being verified 
    verified_dict  =  params.verified_reader.get_data(validity_date, leadtime)
    if (verified_dict is None):
        warnings.warn('Verified dataset unavailable for: ' + str(validity_date) + ' + ' + str(leadtime))

    #compute different verification scores
    if (reference_dict is not None) and (verified_dict is not None):

        for this_domain in params.verif_domains:

            imin,jmin,imax,jmax = params.domain_dict[this_domain]
            #keep only data points in the verification domain
            verif_latitudes    =      reference_dict['lats'][imin:imax,jmin:jmax]
            verif_longitudes   =      reference_dict['lons'][imin:imax,jmin:jmax]
            verif_reference_pr =    reference_dict['values'][imin:imax,jmin:jmax]
            verif_reference_qi = reference_dict['qi_values'][imin:imax,jmin:jmax]
            verif_verified_pr  =     verified_dict['values'][imin:imax,jmin:jmax]
            #quality index < min_qi set to zero
            verif_reference_qi   = np.where(verif_reference_qi < params.min_qi,  0., verif_reference_qi)

            ##TODO REMOVE!!! just for testing
            #verif_reference_qi   = np.where(verif_reference_qi < params.min_qi,  0., 1.)

            #list of points to be used fro corr coeff calculations 
            good_pts = (verif_reference_qi > params.min_qi).nonzero()


            #
            #
            #open sqlite file
            conn = params.sql_handler.connect(must_exist=True)
            c = conn.cursor()

            #id element for the domain being verified
            domain_id = params.sql_handler.domain_id_dict[this_domain]

            #
            #
            # histograms

            #id reference data
            #a new row in sql table is created if it does not already exists
            elem_id, already_existed = params.sql_handler.get_element_id(c, table_name='hist_reference', 
                                                                            domain_id=domain_id, 
                                                                            leadtime_minutes=leadtime_minutes)
            if params.complete_mode == 'complete' and already_existed:
                #in complete mode, we skip element if it already exists
                pass
            else:
                reference_hist, _ = np.histogram(verif_reference_pr, bins=params.hist_bin_edges, weights=verif_reference_qi)
                params.sql_handler.insert_hist(c, 'hist_reference', elem_id, reference_hist, params)
                conn.commit()

            #if of verified data
            #a new row in sql table is created if it does not already exists
            elem_id, already_existed = params.sql_handler.get_element_id(c, table_name='hist_verified', 
                                                                            domain_id=domain_id, 
                                                                            leadtime_minutes=leadtime_minutes)
            if params.complete_mode == 'complete' and already_existed:
                #in complete mode, we skip element if it already exists
                pass
            else:
                verified_hist, _ = np.histogram(verif_verified_pr, bins=params.hist_bin_edges, weights=verif_reference_qi)
                params.sql_handler.insert_hist(c, 'hist_verified', elem_id, verified_hist, params)
                conn.commit()

            conn.close()
            #


            for this_threshold in params.thresholds:

                #id element for the threshold being veridied
                threshold_id = params.sql_handler.threshold_id_dict[str(this_threshold)]

                #a new row in sql table is created if it does not already exists
                conn = params.sql_handler.connect(must_exist=True)
                c = conn.cursor()
                elem_id, already_existed = params.sql_handler.get_element_id(c, table_name='stats', 
                                                                                domain_id=domain_id, 
                                                                                threshold_id=threshold_id, 
                                                                                leadtime_minutes=leadtime_minutes)
                conn.commit()
                conn.close()

                if params.complete_mode == 'complete' and already_existed:
                    #in complete mode, we skip element if it already exists
                    pass
                else:

                    print("verifying date: ", validity_date, "leadtime:", leadtime_minutes, "threshold: ", this_threshold)

                    #apply threshold on quantity
                    reference_true = np.where(verif_reference_pr >= this_threshold, 1., 0.)
                    verified_true  = np.where(verif_verified_pr  >= this_threshold, 1., 0.)

                    #
                    #
                    # contingency table ; data pointa are weighted by quality index
                    x = np.sum(verif_reference_qi * np.logical_and(np.isclose(verified_true,1.), np.isclose(reference_true,1.)))
                    y = np.sum(verif_reference_qi * np.logical_and(np.isclose(verified_true,0.), np.isclose(reference_true,1.)))
                    z = np.sum(verif_reference_qi * np.logical_and(np.isclose(verified_true,1.), np.isclose(reference_true,0.)))
                    w = np.sum(verif_reference_qi * np.logical_and(np.isclose(verified_true,0.), np.isclose(reference_true,0.)))

                    #
                    #
                    # pearson's correlation coefficient
                    reference_pts_for_corr = np.where(verif_reference_pr > this_threshold, verif_reference_pr, 0.)[good_pts]
                    verified_pts_for_corr  = np.where(verif_verified_pr  > this_threshold, verif_verified_pr,  0.)[good_pts]
                    corr_coeff, _ = stats.pearsonr(reference_pts_for_corr, verified_pts_for_corr)

                    #
                    #
                    #minimum length scale for significant results using FSS in circular areas
                    lmin_value = gridded_obs.lmin_from_fss(reference_true, verif_reference_qi, verified_true, 'fft',
                                                           params.lmin_range, grid_dx=params.grid_dx)
                    #for debugging
                    #if lmin_value is None:
                    #    #plot image
                    #    compare_fields.plot_img(reference_true, verif_reference_qi, verified_true, 
                    #                            verif_latitudes,   verif_longitudes,
                    #                            this_threshold, 
                    #                            params.figure_dir, validity_date, leadtime, radius=lmin_value)


                    #
                    #
                    #oper sql file for writting
                    conn = params.sql_handler.connect(must_exist=True)
                    c = conn.cursor()

                    #write x,y,z,w
                    cmd = '''update stats set x = ? , y = ? , z = ? , w = ?  where (id = '''+str(elem_id)+'''); '''
                    c.execute(cmd,(int(x),int(y),int(z),int(w)))

                    #write correlation coefficient
                    cmd = '''update stats set corr_coeff = ? where (id = '''+str(elem_id)+'''); '''
                    c.execute(cmd,(corr_coeff,))

                    #write lmin
                    if lmin_value is not None:
                        cmd = '''update stats set lmin = ? where (id = '''+str(elem_id)+'''); '''
                        c.execute(cmd,(lmin_value,))
                
                    conn.commit()
                    conn.close()


    else:
        #one (or both) of reference or verified fields not found
        pass


    #collect garbage that may accumulate
    gc.collect()

    #return something to maks dask happy
    return np.array([0.],dtype=np.float)



def forecast_loop(params):

    """
    When verifying data, one must loop over:

        - Validity date (T0)
        - Forecast lead time  (only applicable to forecasts, will be 0 when verifying an observation dataset against another)
        - members (only applicable to ensembles)
        - Thresholds
        - Verification domains

    This function is designed for verifying deterministic forecasts

    It loops over Validity dates while leadtimes for a given date are computed in parallel.

    args:
        - params see description in main

    """
    import os
    import time
    import datetime
    import numpy as np
    import dask.distributed
    import dask
    import dask.array as da
    #local
    import gridded_obs.sql 
    import gridded_obs.common


    #list of dates over which to iterate
    date_list = gridded_obs.common.make_date_list(params)

    #list of leadtimes (datetime.timedelta) that will be verified in parallel
    leadtime_list = gridded_obs.common.make_leadtime_list(params)


    #iterate over forecast validity dates
    for date in date_list:

        #for deterministic forecast verification, there is one sqlite file per date verified
        #initialize sqlite file and object to interact with it
        params.sql_handler = gridded_obs.sql.Handler(date, params, params.reference_reader.name, params.verified_reader.name)
        params.sql_handler.init_sqlite_file(params)

        if params.n_cpus == 1:
            #serial execution
            for leadtime in leadtime_list:
                out = verify_date_lt(params, date, leadtime)
        else:
            #parallel executions
            with dask.distributed.Client(processes=True, threads_per_worker=1, n_workers=params.n_cpus, silence_logs=40) as client:

                # sample output so dask knows what to expect only shape and type matters here
                sample = np.array([0],dtype=np.float)  
                #list of desired outputs
                delayed_params = dask.delayed(params)
                res_list = [da.from_delayed(dask_verify_date_lt(delayed_params, date, leadtime), sample.shape, sample.dtype) for leadtime in leadtime_list]
                res_array = da.stack(res_list,axis=1)
                # expected result
                res = res_array.sum()
                # Computation is performed here
                t1 = time.time()
                total = res.compute()
                t2 = time.time()
                print("date: ", date, 'completed in: ', t2-t1, 'seconds')


def verify():

    """launch verification


    Verification is performed between a "reference" and a "verified" datasets

    All arguments prefixed with '--reference_' will be passed directly to reader that must read the referece dataset
             "                  '--verified_'                     "

    gridded_obs does not care about what kind of data is being verified. 
    It is the user's responsability to make sure that the reference and verified readers 
    return quantities that are comparable (eg. with the same units).

    """

    import argparse
    import datetime
    import numpy as np
    import dask.distributed
    #local 
    import gridded_obs.readers 
    import gridded_obs.common

    #arguments for the verification
    parser = argparse.ArgumentParser(description="general variables necessary to all verifications", 
                                     prefix_chars='-+', formatter_class=argparse.RawDescriptionHelpFormatter)


    parser.add_argument("--date_0"  ,     type=lambda d: datetime.datetime.strptime(d, '%Y%m%d%H'),   
                                          required=True,  help=("[yyyymmddhh] (inclusive) first date being verified"))

    parser.add_argument("--date_f"  ,     type=lambda d: datetime.datetime.strptime(d, '%Y%m%d%H'),   
                                          required=False, help=("[yyyymmddhh] (inclusive) last date being verified"), 
                                          default=None)

    parser.add_argument("--delta_date",            type=float, required=False,  help="[minutes] interval between dates being verified",
                                                   default=None)

    parser.add_argument("--leadtime_0" ,           type=float, required=False,  help="[minutes] (inclusive) first leadtime to verify ",
                                                   default=0)
    parser.add_argument("--leadtime_f" ,           type=float, required=False,  help="[minutes] (inclusive) last  leadtime to verify ",
                                                   default=0)
    parser.add_argument("--delta_leadtime",        type=float, required=False, help="[minutes] interval between lead times being verified",
                                                   default=None)
    parser.add_argument("--grid_dx"    ,           type=float, required=True,  help="[km] size of grid tile")
    #output locations
    parser.add_argument("--outname_file_struc",    type=str,   required=False,  help="File structure for output name; use '%reference_name' and '%verified_name' ", 
                                                   default='%verified_name_vs_%reference_name__%Y%m%d%H.sqlite3')
    parser.add_argument("--score_dir" ,            type=str,   required=True,  help="directory where scores will be saved for aggregation later")
    parser.add_argument("--figure_dir" ,           type=str,   required=True,  help="directory where figures will be saved")
    #parameters governing the verification itself
    parser.add_argument("--img_dt",                type=float, required=True,  help="[minutes] frequency at which figures will be generated")
    parser.add_argument("--verif_domains",nargs="+",type=str,  required=True,  help="[imin,jmin,imax,jmax] bottom-left and top-right verificatoin domain ")
    parser.add_argument("--thresholds",  nargs="+",type=float, required=True,  help="list of thresholds to verify")
    parser.add_argument("--k_nbins",               type=int,   required=True,  help="Number of bins for DCT spectra")
    parser.add_argument("--min_qi",                type=float, required=True,  help="Minimum quality index to consider for all verification")
    parser.add_argument("--hist_nbins",            type=int,   required=True,  help="Number of bins for histograms")
    parser.add_argument("--hist_min",              type=float, required=True,  help="[in unit being verified] minimum value for histograms ")
    parser.add_argument("--hist_max",              type=float, required=True,  help="[in unit being verified] maximum value for histograms ")
    parser.add_argument("--hist_log_scale",        type=str,   required=True,  help="set to True to use logscale in histograms ")
    parser.add_argument("--lmin_range",  nargs="+",type=float, required=False, help="minimum and maximum [km] radius for searching lmin ")
    parser.add_argument("--n_cpus",                type=int,   required=True,  help="number of cpus for parallel execution set to 1 for serial execution")
    parser.add_argument("--complete_mode",         type=str,   required=False, default='clobber', help="'complete' or 'clobber' existing sql files")

    #parse arguments
    (params, unknown_args) = parser.parse_known_args()


    #convert strings to bool for certain variables
    params.hist_log_scale = gridded_obs.common.str2bool(params.hist_log_scale)

    #compute edges of histograms
    if params.hist_log_scale:
        params.hist_bin_edges = np.geomspace(params.hist_min, params.hist_max, num=params.hist_nbins+1, endpoint=True)
    else:
        params.hist_bin_edges =  np.linspace(params.hist_min, params.hist_max, num=params.hist_nbins+1, endpoint=True)

    #---------------------------------------------------------------------------


    #TODO 
    #find out a way to pass multiple domains through arguments
    domain_dict = {'all':   [None, None, None, None], 
                   'radars':[200,  50,   2300, 700] }
    params.domain_dict = domain_dict


    #initialize readers for reference data and verified data
    params.reference_reader = gridded_obs.readers.reader_init('reference', unknown_args)
    params.verified_reader  = gridded_obs.readers.reader_init('verified',  unknown_args)


    ##initialize dask client if parallel execution is desired
    #params.dask_client_exists = False
    #if params.n_cpus > 1 :
    #    client = dask.distributed.Client(processes=True, threads_per_worker=1, n_workers=params.n_cpus, silence_logs=40)
    #    params.dask_client_exists = True

    #make sure delta_date is specified
    if params.date_f is not None:
        if params.delta_date is None:
            raise ValueError('argument "delta_date" must be specified when date_f is specified')


    #launch verification
    forecast_loop(params)



if __name__ =='__main__':
    verify()




import dask
@dask.delayed
def dask_verify(*args, **kwargs):
    return verify(*args, **kwargs)

def twod_stats(domain, params):
    """aggregate data for 2D stats
    """

    import datetime
    import warnings
    import numpy as np
    import gridded_obs.sql
    import sqlite3
    import time
    import gridded_obs.plot

    #list of indices to average in each averaging windows
    twod_avg_list = []
    this_lt = params.leadtime_list[0]
    deltat = datetime.timedelta(seconds=params.twod_deltat*60.)
    while this_lt < params.leadtime_list[-1]:
        lower_bound = this_lt
        upper_bound = this_lt + deltat
        bool_list = [ lt >= lower_bound and lt < upper_bound for lt in params.leadtime_list]
        twod_avg_list.append(np.array(bool_list).nonzero()[0])
        this_lt += deltat

    #array of leadtime in minutes
    leadtime_minutes = np.array([ lt.days*1440. + lt.seconds/60. for lt in  params.leadtime_list], dtype=np.int)

    #each averaging window becomes a pannel
    n_windows = len(twod_avg_list)
    n_dates       = len(params.date_list)
    n_experiments = len(params.exp_list)


    #in this loop, aggregation arrays are filled, data for each experiment 
    # are only averaged in the different averaging windows
    ref_bin_edges = None
    for kk, experiment in enumerate(params.exp_list):
        print('')
        print(experiment, '   domain= ', domain)
        for jj, date in enumerate(params.date_list):

            sql_handler = gridded_obs.sql.Handler(date, params, params.reference_name, experiment )
            #print('reading data for '+experiment+str(date))
            conn = sql_handler.connect(none_for_nonexistent=True) 
            if conn is None:
                #file does not exists
                continue
            c = conn.cursor()

            #get histogram bin edges and make sure they are the same for all files being considered 
            hist_bin_edges = sql_handler.get_hist_bin_edges(c)
            if ref_bin_edges is None:
                #
                #this part of the code is run only once for first experiment and date being verified
                #

                #populate reference edges
                ref_bin_edges = hist_bin_edges
                #string that allows to read in histogram data
                hist_string = sql_handler.hist_strings_from_edges(ref_bin_edges, str_type='select')

                #now that we know the dimension of histograms, we initialize result arrays
                n_hist_bins = hist_bin_edges.size - 1
                verified_hists  = np.full((n_experiments, n_windows, n_dates, n_hist_bins), np.nan)
                reference_hists = np.full((               n_windows, n_dates, n_hist_bins), np.nan)

            else:
                if np.sum(np.isclose(ref_bin_edges, hist_bin_edges)) != len(ref_bin_edges) :
                    raise ValueError('hist_bin_edges are not the same for all files encountered.')

            #domain id for this experiment and date
            domain_id = sql_handler.get_domain_id(c, domain)

            #compute average histogram for the different averaging windows
            for ii, lt_inds in enumerate(twod_avg_list):

                #for each leadtime in window
                window_size = len(lt_inds)
                dims = (n_hist_bins, window_size)
                window_verified_hists  = np.full(dims, np.nan)
                window_reference_hists = np.full(dims, np.nan)
                for ee, lt_ind in enumerate(lt_inds):
                    
                    verified_hist_vals  = sql_handler.get_hist_values(c, 'hist_verified', leadtime_minutes[lt_ind], 
                                                                        domain_id, hist_bin_edges, hist_string)
                    reference_hist_vals = sql_handler.get_hist_values(c, 'hist_reference', leadtime_minutes[lt_ind], 
                                                                         domain_id, hist_bin_edges, hist_string)
                    if reference_hist_vals is None or verified_hist_vals is None :
                        continue
                    else:
                        window_verified_hists[:,ee]  = verified_hist_vals
                        window_reference_hists[:,ee] = reference_hist_vals

                #record averaged histogram foor this windo and experiment
                verified_hists[kk, ii, jj, :] = np.nanmean(window_verified_hists, axis=1)
                if kk == 0:
                    reference_hists[ii, jj, :] = np.nanmean(window_reference_hists, axis=1)

            #we are done with this sql file
            conn.close()

    #
    #
    #Here, we make sure that what shows up as NaN in one experiment gets set to Nan in all experiments
    if params.make_same:
        #Make NaN in one experiment NaN in all experiments
        #we flatten 3D arrays along the experiment direction (axis=2) to propagate where there are NaN in any experiment
        v_hist_nan = np.logical_not(np.isfinite(np.sum(verified_hists, axis=0)))
        r_hist_nan = np.logical_not(np.isfinite(reference_hists))
        hist_nan_inds = np.logical_or(v_hist_nan, r_hist_nan).nonzero()

        #propagate NaNs to reference
        #an entry missing in one of experiment or reference is marked as missing for all experiments
        reference_hists[hist_nan_inds] = np.nan
        for kk, experiment in enumerate(params.exp_list):
            verified_hists[kk,:,:,:][hist_nan_inds] = np.nan

    #
    #
    #in this loop, we average data for all experiments
    avg_reference_hists = np.nanmean(reference_hists, axis=1)
    avg_verified_hists  = np.nanmean(verified_hists, axis=2)

    #
    #
    #make figure
    gridded_obs.plot.twod_panels(params, domain, twod_avg_list, leadtime_minutes,
                                 hist_bin_edges, avg_reference_hists, avg_verified_hists)



def time_series(td, params):
    """aggregate data for time series
    """

    import warnings
    import numpy as np
    import gridded_obs.sql
    import sqlite3
    import time
    import gridded_obs.plot

    #threshold and domain verified in this functions call
    threshold = td[0]
    domain = td[1]

    #dimension of aggregation array
    n_leadtimes   = len(params.leadtime_list)
    n_dates       = len(params.date_list)
    n_experiments = len(params.exp_list)

    #initialize aggregation arrays
    dims = (n_leadtimes, n_dates, n_experiments)
    x             = np.full(dims, np.nan)
    y             = np.full(dims, np.nan)
    z             = np.full(dims, np.nan)
    w             = np.full(dims, np.nan)
    lmin          = np.full(dims, np.nan)
    corr_coeff    = np.full(dims, np.nan)

    #fill aggregation arrays
    for kk, experiment in enumerate(params.exp_list):
        print('')
        print(experiment, '   threshold= ',threshold, '   domain= ', domain)
        for jj, date in enumerate(params.date_list):
            sql_handler = gridded_obs.sql.Handler(date, params, params.reference_name, experiment )
            #print('reading data for '+experiment+str(date))
            conn = sql_handler.connect(none_for_nonexistent=True) 
            if conn is None:
                #file does not exists
                continue
            c = conn.cursor()
            c.row_factory = sqlite3.Row
            domain_id = sql_handler.get_domain_id(c, domain)
            if domain_id is None :
                continue
            threshold_id = sql_handler.get_threshold_id(c, threshold)
            if threshold_id is None:
                continue
            cmd = '''select leadtime_minutes, x, y, z, w, lmin, corr_coeff
                                              from stats where (domain_id    = '''+str(domain_id)+''' and 
                                                                threshold_id = '''+str(threshold_id)+''' ) ;''' 
            c.execute(cmd)
            entries = c.fetchall()
            conn.close()

            for entry in entries:
                #leadtime index
                ii = int((entry['leadtime_minutes'] - params.leadtime_0) / params.delta_leadtime)
                x[ii,jj,kk]          = entry['x']
                y[ii,jj,kk]          = entry['y']
                z[ii,jj,kk]          = entry['z']
                w[ii,jj,kk]          = entry['w']
                lmin[ii,jj,kk]       = entry['lmin']
                corr_coeff[ii,jj,kk] = entry['corr_coeff']

        #we use finite number of hits to identify experiment with no data whatsoever
        #this will cause everything else to be set to NaN
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            if np.isclose(np.nansum(x[:,:,kk]), 0.):
                raise RuntimeError('Experiment:'+experiment+' has no data at all. We stop here as this would cause everything else to be rejected.')

    if params.make_same:
        #Make NaN in one experiment NaN in all experiments
        #we flatten 3D arrays along the experiment direction (axis=2) to propagate where there are NaN in any experiment
        x_nan_inds          = np.logical_not(np.isfinite(np.sum(x,         axis=2))).nonzero()
        y_nan_inds          = np.logical_not(np.isfinite(np.sum(y,         axis=2))).nonzero()
        z_nan_inds          = np.logical_not(np.isfinite(np.sum(z,         axis=2))).nonzero()
        w_nan_inds          = np.logical_not(np.isfinite(np.sum(w,         axis=2))).nonzero()
        lmin_nan_inds       = np.logical_not(np.isfinite(np.sum(lmin,      axis=2))).nonzero()
        corr_coeff_nan_inds = np.logical_not(np.isfinite(np.sum(corr_coeff,axis=2))).nonzero()
        #mark missing for all experiments
        #an entry missing in one experiment is marked as missing for all experiments
        for kk, experiment in enumerate(params.exp_list):
            x[:,:,kk][x_nan_inds]                   = np.nan
            y[:,:,kk][y_nan_inds]                   = np.nan
            z[:,:,kk][z_nan_inds]                   = np.nan
            w[:,:,kk][w_nan_inds]                   = np.nan
            lmin[:,:,kk][lmin_nan_inds]             = np.nan
            corr_coeff[:,:,kk][corr_coeff_nan_inds] = np.nan

    #
    #
    #time_series
    #average statistics
    avg_dims = (n_leadtimes, n_experiments)
    avg_fbias = np.full(avg_dims, np.nan)
    avg_pod   = np.full(avg_dims, np.nan)
    avg_far   = np.full(avg_dims, np.nan)
    avg_csi   = np.full(avg_dims, np.nan)
    avg_gss   = np.full(avg_dims, np.nan)
    avg_lmin  = np.full(avg_dims, np.nan)
    avg_corr_coeff = np.full(avg_dims, np.nan)
    #ignore warnings for /0 we can handle NaN later
    with np.errstate(divide='ignore', invalid='ignore'):
        for kk, experiment in enumerate(params.exp_list):
            #sum x, y, z, w
            x_sum = np.nansum(x[:,:,kk], axis=1)
            y_sum = np.nansum(y[:,:,kk], axis=1)
            z_sum = np.nansum(z[:,:,kk], axis=1)
            w_sum = np.nansum(w[:,:,kk], axis=1)

            #make sure numbers are the same for each category
            nx = np.nansum(np.isfinite(x[:,:,kk]), axis=1)
            ny = np.nansum(np.isfinite(y[:,:,kk]), axis=1)
            nz = np.nansum(np.isfinite(z[:,:,kk]), axis=1)
            nw = np.nansum(np.isfinite(w[:,:,kk]), axis=1)
            if np.sum(nx-ny) > 0. or  np.sum(nx-nz) > 0. or np.sum(nx-nw) > 0.:
                raise ValueError('Number of x,y,x,w being averaged is not the same...')
            else: 
                n_avg_contingency = nx

            avg_fbias[:,kk] =  (x_sum + z_sum)/(x_sum + y_sum)          #frequency bias
            avg_pod[:,kk]   = x_sum / (x_sum + y_sum)                   #probability of detection  % of events that are forecasted
            avg_far[:,kk]   = z_sum/(x_sum + z_sum)                     #false alarm rate        (# of false alarms) / (# of predicted events)
            avg_csi[:,kk]   = x_sum/(x_sum+y_sum+z_sum)                 #critical sucess index    (# of hits) / (# of events  +  # of false alarm)
            c = (x_sum+y_sum)*(x_sum+z_sum)/(x_sum+y_sum+z_sum+w_sum)   #event frequency
            avg_gss[:,kk]   = (x_sum-c)/(x_sum-c+y_sum+z_sum)           #Gilbert skill score      # of correct forecasts in excess of those expected by chance 
                                                                        #       / # of cases when there was a threat that would not have been forecasted by chance
                                                                        # think about it as csi corrected by chance            
                                                                        
            #other scores are simply averaged
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)

                #lmin
                avg_lmin[:,kk]        = np.nanmean(lmin[:,:,kk], axis=1)
                n_avg_lmin = np.nansum(np.isfinite(lmin[:,:,kk]), axis=1)

                #corr coeff
                avg_corr_coeff[:,kk] = np.nanmean(corr_coeff[:,:,kk], axis=1)
                n_avg_corr_coeff  = np.nansum(np.isfinite(corr_coeff[:,:,kk]), axis=1)

            #force certain leadtimes to NaN if the user wants this
            if params.leadtime_ignore_inds is not None:
                lt_inds = params.leadtime_ignore_inds
                avg_fbias[lt_inds,kk]      = np.nan
                avg_pod[lt_inds,kk]        = np.nan
                avg_far[lt_inds,kk]        = np.nan
                avg_csi[lt_inds,kk]        = np.nan
                avg_gss[lt_inds,kk]        = np.nan
                avg_lmin[lt_inds,kk]       = np.nan
                avg_corr_coeff[lt_inds,kk] = np.nan

    score_dict = {'fbias':     avg_fbias,
                  'pod':       avg_pod,  
                  'far':       avg_far,  
                  'csi':       avg_csi,  
                  'gss':       avg_gss,  
                  'lmin':      avg_lmin,
                  'corr_coeff':avg_corr_coeff, 
                  #
                  'n_fbias':     n_avg_contingency,
                  'n_pod':       n_avg_contingency,
                  'n_far':       n_avg_contingency,
                  'n_csi':       n_avg_contingency,
                  'n_gss':       n_avg_contingency,
                  'n_lmin':      n_avg_lmin,
                  'n_corr_coeff':n_avg_corr_coeff }

    #plot average error statistics 
    gridded_obs.plot.time_series(params, threshold, domain, score_dict)


def aggregate():

    """launch aggregation of pre-computed scores

    Several experiment can be compared side by side

    Checks make sure that the same dataset and verification parameters were 
    used for each experiment being compared

    Two types of graphs are produced:
        - time series
        - 2d plot in time intervals

    In all cases special care is taken that the same observations be used for all experiments

    """

    import argparse
    import datetime
    import numpy as np
    import dask.distributed
    import gridded_obs.common
    import domutils.legs as legs

    #arguments for the verification
    parser = argparse.ArgumentParser(description="variables necessary for score aggregation", 
                                     prefix_chars='-+', formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("--date_0"  ,     type=lambda d: datetime.datetime.strptime(d, '%Y%m%d%H'),   
                                          required=True,  help=("[yyyymmddhh] (inclusive) first date being verified"))

    parser.add_argument("--date_f"  ,     type=lambda d: datetime.datetime.strptime(d, '%Y%m%d%H'),   
                                          required=False, help=("[yyyymmddhh] (inclusive) last date being verified"), 
                                          default=None)

    parser.add_argument("--delta_date",            type=float, required=False,  help="[minutes] interval between dates being verified",
                                                   default=None)

    parser.add_argument("--leadtime_0" ,           type=float, required=False,  help="[minutes] (inclusive) first leadtime to verify ",
                                                   default=0)
    parser.add_argument("--leadtime_f" ,           type=float, required=False,  help="[minutes] (inclusive) last  leadtime to verify ",
                                                   default=0)
    parser.add_argument("--delta_leadtime",        type=float, required=False, help="[minutes] interval between lead times being verified",
                                                   default=None)
    parser.add_argument("--leadtime_ignore",       nargs="+",type=float, required=False,  help="[minutes] list of leadtimes that will be set to NaN",
                                                   default=None)
    parser.add_argument("--leadtime_greyed_out",   nargs="+",type=float, required=False,  help="[minutes] min and max leadtime to be greyed out",
                                                   default=None)
    parser.add_argument("--make_same",             type=str, required=False, help="make NaN in one experiment NaN in all experiments", default=True)

    #what are we to plot
    parser.add_argument("--time_series",   nargs="+",type=str,  required=False, default=None, help="quantities to plot, from the bottom up, in time_series figures")
    parser.add_argument("--twod_panels",   nargs="+",type=str,  required=False, default=None, help="quantities to plot, from the bottom up, in 2D figures")
    parser.add_argument("--twod_deltat",             type=float,required=False, default=60., help="[minutes] time interval for 2D figures")


    #yrange for figures to be plotted   if no arguments passed, range is automatic
    parser.add_argument("--ylim_fbias",      nargs="+", type=float, required=False, help="min, max ;  yrange for fbias plots",      default=None)
    parser.add_argument("--ylim_pod",        nargs="+", type=float, required=False, help="min, max ;  yrange for pod plots",        default=None)
    parser.add_argument("--ylim_far",        nargs="+", type=float, required=False, help="min, max ;  yrange for far plots",        default=None)
    parser.add_argument("--ylim_csi",        nargs="+", type=float, required=False, help="min, max ;  yrange for csi plots",        default=None)
    parser.add_argument("--ylim_lmin",       nargs="+", type=float, required=False, help="min, max ;  yrange for lmin plots",       default=None)
    parser.add_argument("--ylim_corr_coeff", nargs="+", type=float, required=False, help="min, max ;  yrange for corr_coeff plots", default=None)

    #output locations
    parser.add_argument("--score_dir" ,              type=str,   required=True,  help="directory where scores will be saved for aggregation later")
    parser.add_argument("--figure_dir" ,             type=str,   required=True,  help="directory where figures will be saved")
    #parameters governing the aggregation
    parser.add_argument("--verif_domains", nargs="+",type=str,   required=True,  help="[imin,jmin,imax,jmax] bottom-left and top-right verificatoin domain ")
    parser.add_argument("--thresholds",    nargs="+",type=float, required=True,  help="list of thresholds to verify")

    parser.add_argument("--n_cpus",                  type=int,   required=True,  help="number of cpus for parallel execution set to 1 for serial execution")

    #parameters governing the experiments being verified
    parser.add_argument("--reference_name",          type=str,  required=True,  help="name of reference dataset for all experiments being compared")
    parser.add_argument("--exp_list",     nargs="+", type=str,  required=True,  help="list of experiments to plot")
    parser.add_argument("--exp_color",    nargs="+", type=str,  required=False, default=None,  help="list of colors for each experiment see comment above")
    parser.add_argument("--exp_linestyle",nargs="+", type=str,  required=False, default=None,  help="matplotlib linestyle for each experiment")
    parser.add_argument("--exp_linewidth",nargs="+", type=float,required=False, default=None,  help="iatplotlib linewidth for each experiment")

    #parse arguments
    (params, unknown_args) = parser.parse_known_args()

    if len(unknown_args) > 0:
        print('unknown arguments:')
        print(unknown_args)

    #---------------------------------------------------------------------------

    #convert strings to bool for certain variables
    params.make_same = gridded_obs.common.str2bool(params.make_same)

    #initialize dask client if parallel execution is desired
    params.dask_client_exists = False
    if params.n_cpus > 1 :
        client = dask.distributed.Client(processes=True, threads_per_worker=1, n_workers=params.n_cpus, silence_logs=40)
        params.dask_client_exists = True

    #make sure delta_date is specified
    if params.date_f is not None:
        if params.delta_date is None:
            raise ValueError('argument "delta_date" must be specified when date_f is specified')

    #default if unspecified
    if params.exp_color is None:
        default_colors = ['l_blue_0.7','l_green_0.7','l_orange_0.7','l_red_0.7','l_pink_0.7','l_purple_0.7','l_yellow_0.7','l_brown_0.7']
        params.exp_color = default_colors[0:len(params.exp_list)]

    if params.exp_linestyle is None:
        params.exp_linestyle = 'solid' * len(params.exp_list)

    if params.exp_linewidth is None:
        params.exp_linewidth = [10] * len(params.exp_list)

    #validate number of arguments

    if len(params.exp_color) != len(params.exp_list):
        raise ValueError('exp_color must have the same number of elements ar exp_list')

    if len(params.exp_linestyle) != len(params.exp_list):
        raise ValueError('exp_linestyle must have the same number of elements ar exp_list')

    if len(params.exp_linewidth) != len(params.exp_list):
        raise ValueError('exp_linewidth must have the same number of elements ar exp_list')

    #process colors into something that matplotlib will like
    for ii, input_color in enumerate(params.exp_color):
        if input_color[0] == '(' and input_color[-1] == ')':
            #we have (r,g,b) or (r,g,b,a)
            rgb_list = input_color[1:-1].split(',')
            if len(rgb_list) == 3:
                #we have (r,g,b) 
                params.exp_color[ii] = tuple(np.array([np.float(elem)/255. for elem in rgb_list]))
            if len(rgb_list) == 4:
                #we have (r,g,b,a) 
                rgb_arr = np.array([np.float(elem)/255. for elem in rgb_list])
                rgb_arr[-1] = np.float(rgb_list[-1])  #   alpha should not be divided by 255.
                params.exp_color[ii] = tuple(rgb_arr)
            else:
                raise ValueError("color '(r,g,b)' or '(r,g,b,a)' can only have 3 or 4 elements")
        elif input_color[0:2] == 'l_':
            #we have a legs color of type:  'l_orange_0.7'
            try:
                color, value = input_color[2:].split('_')
            except:
                raise ValueError("conversion problem: legs color should be of the form 'l_orange_0.7")
            c_map = legs.PalObj(color_arr=color)
            rgb_arr = c_map.to_rgb(value)[0]/255.
            params.exp_color[ii] = tuple(rgb_arr)
        else:
            #in all other cases, the string present in params.exp_color will be passed directly to matplotlib
            pass

    #list of dates over which to iterate
    params.date_list = gridded_obs.common.make_date_list(params)

    #list of leadtimes (datetime.timedelta) that will be verified in parallel
    params.leadtime_list = gridded_obs.common.make_leadtime_list(params)

    #list of indexes of leadtime that should be ignored
    params.leadtime_ignore_inds = None
    if params.leadtime_ignore is not None:
        inds = []
        for leadtime_m in params.leadtime_ignore:
            leadtime_dt = datetime.timedelta(seconds=leadtime_m*60.)
            ind = np.array([lt == leadtime_dt for lt in params.leadtime_list]).nonzero()[0][0]
            inds.append(ind)
        params.leadtime_ignore_inds = inds



    #
    #
    #2D plots
    #domains can be computed in parallel
    for domain in params.verif_domains:
        twod_stats(domain, params)

    #
    #
    #time series
    #domain/thresholds can be computed in parallel
    td_list = [(threshold,domain) for threshold in params.thresholds for domain in params.verif_domains]
    for td in td_list:
        time_series(td, params)



if __name__ =='__main__':
    aggregate()







def time_series(params, threshold, domain, score_dict):
    """plot time series

    what is being plotted is defined in params.time_series from the bottom up
    """

    import numpy as np
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import domcmc.fst_tools as fst_tools
    import domutils.legs as legs
    import domutils.geo_tools as geo_tools
    import domutils.radar_tools as radar_tools
    import domutils._py_tools as py_tools
    import gridded_obs.plot

    n_panels  = len(params.time_series)

    #missing val
    missing = -9999.

    #leadtime being verified
    leadtime_hours = np.array([ lt.days*24. + lt.seconds/3600. for lt in params.leadtime_list])
    xticks = np.arange(-3., 13.,3.)

    #figure dimensions
    ratio  = 0.15
    rec_w  = 10.           # Horizontal size of a panel
    rec_h  = ratio*rec_w  # Vertical size of a panel
    sp_w   = 1.2           # horizontal space between panels
    sp_h   = .2           # vertical space between panels
    pal_sp = .05          # space between panel and palette
    pal_w  = .17          # width of palette
    tit_h  = 1.6 + 0.5*np.ceil(len(params.exp_list)/2.)  # height of title
    #size of figure
    fig_w  = rec_w + 2. * sp_w
    fig_h  = sp_h + (n_panels)*(rec_h+sp_h) + tit_h + .9
    #normalize sizes relative to figure
    rec_w  = rec_w / fig_w
    rec_h  = rec_h / fig_h
    sp_w   = sp_w / fig_w
    sp_h   = sp_h / fig_h
    pal_sp = pal_sp / fig_w
    pal_w  = pal_w / fig_w
    tit_h  = tit_h / fig_h
    #pixel resolution of image being generated
    grid_w_pts = 1200.
    image_dpi = grid_w_pts/(rec_w*fig_w)
    image_res = [grid_w_pts,ratio*grid_w_pts]

    #larger typeface
    mpl.rcParams.update({'font.size': 20})
    # Use this for editable text in svg
    mpl.rcParams['text.usetex']  = False
    mpl.rcParams['svg.fonttype'] = 'none'
    # Hi def figure
    mpl.rcParams['figure.dpi'] = 600

    #dictionary for vertical range of pannels
    r_dict = {}
    if params.ylim_pod is not None:
        r_dict['pod'] = params.ylim_pod
    else:
        r_dict['pod'] = gridded_obs.plot.autorange(score_dict['pod'])
    if params.ylim_far is not None:
        r_dict['far'] = params.ylim_far
    else:
        r_dict['far'] =  gridded_obs.plot.autorange(score_dict['far'])
    if params.ylim_csi is not None:
        r_dict['csi'] = params.ylim_csi
    else:
        r_dict['csi'] =  gridded_obs.plot.autorange(score_dict['csi'])
    if params.ylim_fbias is not None:
        r_dict['fbias'] = params.ylim_fbias
    else:
        r_dict['fbias'] =  gridded_obs.plot.autorange(score_dict['fbias'])
    if params.ylim_lmin is not None:
        r_dict['lmin'] = params.ylim_lmin
    else:
        r_dict['lmin'] =  gridded_obs.plot.autorange(score_dict['lmin'])
    if params.ylim_corr_coeff is not None:
        r_dict['corr_coeff'] = params.ylim_corr_coeff
    else:
        r_dict['corr_coeff'] =  gridded_obs.plot.autorange(score_dict['corr_coeff'])

    unit_dict = {'fbias':'unitless', 
                 'pod':'unitless', 
                 'far':'unitless', 
                 'csi':'unitless', 
                 'lmin':'radius [km]',
                 'corr_coeff':'unitless'}

    #instantiate figure
    fig = plt.figure(figsize=(fig_w, fig_h))

    #header, legend and other explanations
    gridded_obs.plot.plot_header(fig, domain, params,
                                 fig_w  , fig_h  ,
                                 rec_w  , rec_h,
                                 sp_w   , sp_h,
                                 pal_sp , pal_w,
                                 tit_h, threshold=threshold  )


    x0 = sp_w
    for nn, score in enumerate(params.time_series):
        #plot panels one by one
        y0 = sp_h + nn*(rec_h + sp_h) + 1./fig_h

        pos = [x0, y0, rec_w, rec_h]

        y_lim = r_dict[score]

        #Setup the general look 
        #Ticks length and width
        major_height = 5
        major_width = .3
        minor_height = 2
        minor_width = .3
        #axes containing data
        ax = fig.add_axes(pos, label = 'time_series', zorder=10)
        ax.set_facecolor((0.,0.,0.,0.))
        ax.set_xlim((leadtime_hours[0], leadtime_hours[-1]))
        ax.set_ylim(y_lim)
        ax.tick_params(axis='both', which='both', direction='in', top='on', right=True)
        ax.set_xticks(xticks)
        #ax.set_yticks(this_var_info['ticks']) 
        ax.tick_params('both', length=major_height, width=major_width, which='major')
        ax.tick_params('both', length=minor_height, width=minor_width, which='minor')
        if nn == 0:
            ax.set_xlabel('Leadtime (h)')
        else:
            ax.set_xticklabels([])
        ax.set_ylabel(unit_dict[score])
        #ax.yaxis.set_minor_locator(plt.MultipleLocator(this_var_info['minor_tick_separation']))

        #horizontal line for frequency bias
        if score == 'fbias':
            ax.axhline(1., color='black', zorder=1, linewidth=0.5)

        #thinner lines for axis, god is this syntax ugly...
        for axis in ['top','bottom','left','right']:
              ax.spines[axis].set_linewidth(0.3)

        #greyed out zone in panels
        if params.leadtime_greyed_out is not None:
            gc = 220./255.
            g_xmin = params.leadtime_greyed_out[0]/60.
            g_xmax = params.leadtime_greyed_out[1]/60.
            g_width = g_xmax - g_xmin
            g_height = y_lim[1] - y_lim[0]
            #        y,         width,     height,     left
            ax.barh([y_lim[0]], [g_width], [g_height], left=g_xmin, align='edge', color=(gc,gc,gc,1.), edgecolor=(1.,1.,1.,0.), zorder=0)
    
        #plot name of score on panel
        ax.annotate(score.upper(), (0.02, 0.8), clip_on=False, xycoords='axes fraction',\
                    ha='left', va='center', fontsize=28)
        print('minmax: ',score, np.nanmin(score_dict[score]), np.nanmax(score_dict[score]))


        #plot number of data points in each bin
        x_extent = leadtime_hours[-1] - leadtime_hours[0]
        y_extent = y_lim[1] - y_lim[0]
        #y position of numbers 
        yp = y_lim[0] + 0.95 * y_extent
        #30 number are legible
        mod_num = 1.
        while len(leadtime_hours)/mod_num > 30.:
            mod_num *= 2.
        #print numbers
        for nn, (leadtime, count) in enumerate(zip(leadtime_hours, score_dict['n_'+score])):
            if np.mod(nn,mod_num):
                #skip some entries to make the whole thing legible
                continue
            elif (nn == 0):
                continue
            else:
                ax.text(leadtime, yp, '{:d}'.format(np.int(count)), fontsize=9, color='black', 
                        ha='center', va = 'center', alpha = 0.3)





        #plot time series for each experiments
        for kk, experiment in enumerate(params.exp_list):
            ax.plot(   leadtime_hours, score_dict[score][:,kk], color=params.exp_color[kk], 
                                                                linestyle=params.exp_linestyle[kk], 
                                                                 linewidth=params.exp_linewidth[kk], )

    #save image
    date_dir = params.score_dir+ params.date_list[0].strftime('%Y%m%d%H')+'_to_'+ params.date_list[-1].strftime('%Y%m%d%H')
    py_tools.parallel_mkdir(date_dir)
    svg_name = date_dir+'/time_series__'+domain+'__{:09.4f}'.format(threshold)+'.svg'
    plt.savefig(svg_name, dpi=image_dpi)
    plt.close(fig)
    py_tools.lmroman(svg_name)
    #py_tools.convert(svg_name,'gif', del_orig=True, density=400, geometry='50%')
    print('done with: ', svg_name)






def plot_circles(ax, lats, lons, radius):

    #plot circles to debug lmin code
    import cartopy
    import cartopy.crs 
    import domutils.geo_tools as geo_tools
    import numpy as np

    #plot radar circles  
    proj_cart = cartopy.crs.PlateCarree()
    transform = proj_cart._as_mpl_transform(ax)
    azimuths = np.arange(0.,361)
    for this_lat, this_lon in zip(lats.ravel(),lons.ravel()):
        ranges   = np.full_like(azimuths, radius)
        lon1_arr = np.full_like(azimuths, this_lon)
        lat1_arr = np.full_like(azimuths, this_lat)
        rlons, rlats = geo_tools.lat_lon_range_az(lon1_in   =   lon1_arr, 
                                                  lat1_in   =   lat1_arr,
                                                  range_in  =   ranges,
                                                  azimuth_in=   azimuths)
        #circles
        color=(0./256.,81./256.,237./256.)
        ax.plot(rlons, rlats, transform=proj_cart, c=color, zorder=300, linewidth=.1)
        ax.scatter(this_lon, this_lat, transform=proj_cart, color=color, zorder=300, s=1.**2.)

        break



def plot_geo(ax):
    import cartopy.feature as cfeature
    ax.outline_patch.set_linewidth(.3)
    ax.add_feature(cfeature.STATES.with_scale('50m'), linewidth=0.2, edgecolor='0.6')


def compare_fields(reference_v_in, reference_qi_in, verified_v_in,
                   latitudes, longitudes, 
                   threshold, 
                   pic_dir, validity_date, leadtime, 
                   radius=None):
    """image of reference,quality index and verified fields
    """

    import datetime
    import os
    import numpy as np
    import pickle
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import domcmc.fst_tools as fst_tools
    import domutils.legs as legs
    import domutils.geo_tools as geo_tools
    import domutils.radar_tools as radar_tools
    import domutils._py_tools as py_tools


    #show data only above threshold
    reference_v  = reference_v_in 
    reference_qi = reference_qi_in 
    verified_v  = verified_v_in  


    #date that will be displayed
    this_date = validity_date + leadtime

    #missing val
    missing = -9999.

    #figure dimensions
    ratio  = 0.5
    rec_w  = 6.           # Horizontal size of a panel
    rec_h  = ratio*rec_w  # Vertical size of a panel
    sp_w   = .2           # horizontal space between panels
    sp_h   = 1.           # vertical space between panels
    pal_sp = .05           # spavce between panel and palette
    pal_w  = .17          # width of palette
    tit_h  = .1           # height of title
    #size of figure
    n_exp  = 4
    n_panels = 2
    fig_w  = 21.
    fig_h  = 9.
    #normalize sizes relative to figure
    rec_w  = rec_w / fig_w
    rec_h  = rec_h / fig_h
    sp_w   = sp_w / fig_w
    sp_h   = sp_h / fig_h
    pal_sp = pal_sp / fig_w
    pal_w  = pal_w / fig_w
    tit_h  = tit_h / fig_h
    #pixel resolution of image being generated
    grid_w_pts = 1200.
    image_dpi = grid_w_pts/(rec_w*fig_w)
    image_res = [grid_w_pts,ratio*grid_w_pts]
    #objects to handle color mappings
    pr_color_map = legs.PalObj(range_arr=[.1,.5,1.,5.,10.,20.,50.],
                                  n_col=6, 
                                  over_high='extend', under_low='white',
                                  excep_val=missing, excep_col='grey_220')
    #larger typeface
    mpl.rcParams.update({'font.size': 18})
    # Use this for editable text in svg
    mpl.rcParams['text.usetex']  = False
    mpl.rcParams['svg.fonttype'] = 'none'
    # Hi def figure
    mpl.rcParams['figure.dpi'] = 600

    #custom pastel color segments for QI index
    pastel = [ [[255,190,187],[230,104, 96]],  #pale/dark red
               [[255,185,255],[147, 78,172]],  #pale/dark purple
               [[255,227,215],[205,144, 73]],  #pale/dark brown
               [[210,235,255],[ 58,134,237]],  #pale/dark blue
               [[223,255,232],[ 61,189, 63]] ] #pale/dark green
    qi_color_map = legs.PalObj(range_arr=[0., 1.],
                               dark_pos='high',
                               color_arr=pastel,
                               excep_val=[missing, 0.],
                               excep_col=['grey_220', 'white'])

    # Full national grid
    # Use rotated pole projection
    # Use rotated pole projection
    pole_latitude=35.7
    pole_longitude=65.5
    lat_0 = 48.8
    delta_lat = 9.6
    lon_0 = 266.1
    delta_lon = 40.1
    map_extent_full=[lon_0-delta_lon, lon_0+delta_lon, lat_0-delta_lat, lat_0+delta_lat]  
    crs_full = ccrs.RotatedPole(pole_latitude=pole_latitude, pole_longitude=pole_longitude)

    #instantiate objects to handle geographical projection of data this needs to be done only once
    # onto geoAxes with this specific crs 
    proj_obj_full = geo_tools.ProjInds(src_lon=longitudes, src_lat=latitudes, 
                                       dest_crs=crs_full, extent=map_extent_full,
                                       image_res=image_res)

    #instantiate figure
    fig = plt.figure(figsize=(fig_w, fig_h))

    proj_obj_list = [proj_obj_full]
    crs_list      = [crs_full]
    extent_list   = [map_extent_full]

    for ii, [this_proj_obj, this_crs, this_extent] in enumerate(zip(proj_obj_list,crs_list,extent_list)):

        #Y position of row
        y0 = sp_h + (1.-ii)*(sp_h+rec_h)
        #title offset
        xp = .04
        yp = 1.01

        #reference values
        x0 = 0.1/fig_w
        pos = [x0, y0, rec_w, rec_h]
        ax = fig.add_axes(pos, projection=this_crs, extent=this_extent)
        #title
        ax.annotate('reference_val', size=22, xy=(xp, yp), xycoords='axes fraction')
        #print date only once
        if ii ==0:
            ax.annotate(this_date.strftime('%Y-%m-%d %Hh%M'), size=24, xy=(xp, 1.2), xycoords='axes fraction')
        #geographical projection of data into axes space
        #ddp
        projected_data = this_proj_obj.project_data(reference_v)
        #draw color figure onto axes
        pr_color_map.plot_data(ax=ax, data=projected_data)
        #format axes and draw geographical boundaries
        plot_geo(ax)

        #verified_val
        x0 = 0.1/fig_w + sp_w + rec_w
        pos = [x0, y0, rec_w, rec_h]
        ax2 = fig.add_axes(pos, projection=this_crs, extent=this_extent)
        #title
        ax2.annotate('verified_val', size=22, xy=(xp, yp), xycoords='axes fraction')
        #geographical projection of data into axes space
        #ddp
        projected_data = this_proj_obj.project_data(verified_v)
        #draw color figure onto axes
        pr_color_map.plot_data(ax=ax2, data=projected_data)
        #format axes and draw geographical boundaries
        plot_geo(ax2)
        #palette
        pal_pos = [x0+rec_w+pal_sp, y0, pal_w, rec_h]
        pr_color_map.plot_palette(pal_pos=pal_pos, 
                                     pal_linewidth=0.3, pal_units='[mm]',
                                     pal_format='{:3.0f}', equal_legs=True)

        #
        #
        #plot circles
        if radius is not None:
            delta_pt = 40
            subset_lat =  latitudes[::delta_pt,::delta_pt]
            subset_lon = longitudes[::delta_pt,::delta_pt]
            plot_circles(ax2, subset_lat, subset_lon, radius)

        ##plot circle selection
        #p_file = '/space/hall3/sitestore/eccc/mrd/rpndat/dja001/lmin_investigate/pts.pickle'
        #with open(p_file, 'rb') as f_handle:
        #    rec_dict = pickle.load(f_handle)
        #p_file = '/space/hall3/sitestore/eccc/mrd/rpndat/dja001/lmin_investigate/ll.pickle'
        #with open(p_file, 'rb') as f_handle:
        #    ll_dict = pickle.load(f_handle)
        #proj_cart = ccrs.PlateCarree()
        #print('number of balls', len(rec_dict['pt_list']))
        #for ball_list in rec_dict['pt_list']:
        #    ax2.scatter(ll_dict['longitudes'][ball_list], ll_dict['latitudes'][ball_list], 
        #                transform=proj_cart, color=(203./256.,26./256.,35./256.),zorder=300, s=.05**2.)




        #reference quality index
        x0 = 0.1/fig_w + 2.*(sp_w + rec_w)+ 5.*sp_w
        pos = [x0, y0, rec_w, rec_h]
        ax3 = fig.add_axes(pos, projection=this_crs, extent=this_extent)
        #title
        ax3.annotate('Quality index', size=22, xy=(xp, yp), xycoords='axes fraction')
        #geographical projection of data into axes space
        #ddp
        projected_data = this_proj_obj.project_data(reference_qi)
        #draw color figure onto axes
        qi_color_map.plot_data(ax=ax3, data=projected_data)
        #format axes and draw geographical boundaries
        plot_geo(ax3)
        #palette
        pal_pos = [x0+rec_w+pal_sp, y0, pal_w, rec_h]
        qi_color_map.plot_palette(pal_pos=pal_pos, 
                                  pal_linewidth=0.3, pal_units='[unitless]',
                                  pal_format='{:2.1f}')

    #make dir if it does not exist
    if not os.path.isdir(pic_dir):
        domutils._py_tools.parallel_mkdir(pic_dir)

    #figure name
    if not np.isclose(leadtime.seconds, 0):
        lt_minutes = leadtime.days*1440. + np.floor(leadtime.seconds/60.)
        svg_name = pic_dir+'/compare_fields_'+validity_date.strftime('%Y%m%d%H%M')+'{:+06.0f}m_th{:04.1f}'.format(lt_minutes, threshold)+'.svg'
    else:
        svg_name = pic_dir+'/compare_fields_'+validity_date.strftime('%Y%m%d%H%M')+'_th{:03.0f}'.format(threshold)+'.svg'
    plt.savefig(svg_name, dpi=image_dpi)
    plt.close(fig)
    py_tools.lmroman(svg_name)
    #py_tools.convert(svg_name,'gif', del_orig=True, density=400, geometry='50%')
    print('done with: ', svg_name)


def plot_fraction(reference_v_in, longitudes, latitudes, 
                  rf=None, rfi=None,
                  vf=None, vfi=None,
                  radius=None):
    """image of fractions for FSS
    """

    if rf is not None:
        reference_fractions = rf
    if rfi is not None:
        reference_fractions_int = rfi
    if vf is not None:
        verified_fractions = vf
    if vfi is not None:
        verified_fractions_int = vfi


    import datetime
    import os
    import numpy as np
    import pickle
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import domcmc.fst_tools as fst_tools
    import domutils.legs as legs
    import domutils.geo_tools as geo_tools
    import domutils.radar_tools as radar_tools
    import domutils._py_tools as py_tools


    #show data only above threshold
    reference_v  = reference_v_in

    #missing val
    missing = -9999.

    #figure dimensions
    ratio  = 0.5
    rec_w  = 6.           # Horizontal size of a panel
    rec_h  = ratio*rec_w  # Vertical size of a panel
    sp_w   = .2           # horizontal space between panels
    sp_h   = 1.           # vertical space between panels
    pal_sp = .05           # spavce between panel and palette
    pal_w  = .17          # width of palette
    tit_h  = .1           # height of title
    #size of figure
    n_exp  = 4
    n_panels = 2
    fig_w  = 21.
    fig_h  = 9.
    #normalize sizes relative to figure
    rec_w  = rec_w / fig_w
    rec_h  = rec_h / fig_h
    sp_w   = sp_w / fig_w
    sp_h   = sp_h / fig_h
    pal_sp = pal_sp / fig_w
    pal_w  = pal_w / fig_w
    tit_h  = tit_h / fig_h
    #pixel resolution of image being generated
    grid_w_pts = 1200.
    image_dpi = grid_w_pts/(rec_w*fig_w)
    image_res = [grid_w_pts,ratio*grid_w_pts]
    #objects to handle color mappings
    pr_color_map = legs.PalObj(range_arr=[.1,.5,1.,5.,10.,20.,50.],
                                  n_col=6, 
                                  over_high='extend', under_low='white',
                                  excep_val=missing, excep_col='grey_220')
    #larger typeface
    mpl.rcParams.update({'font.size': 18})
    # Use this for editable text in svg
    mpl.rcParams['text.usetex']  = False
    mpl.rcParams['svg.fonttype'] = 'none'
    # Hi def figure
    mpl.rcParams['figure.dpi'] = 600

    #custom pastel color segments for QI index
    pastel = [ [[255,190,187],[230,104, 96]],  #pale/dark red
               [[255,185,255],[147, 78,172]],  #pale/dark purple
               [[255,227,215],[205,144, 73]],  #pale/dark brown
               [[210,235,255],[ 58,134,237]],  #pale/dark blue
               [[223,255,232],[ 61,189, 63]] ] #pale/dark green
    qi_color_map = legs.PalObj(range_arr=[0., .2],
                               dark_pos='high',
                               color_arr=pastel,
                               over_under='extend',
                               excep_val=[missing, 0.,-1.],
                               excep_col=['grey_220', 'white', 'dark_red'])

    # Full national grid
    # Use rotated pole projection
    # Use rotated pole projection
    pole_latitude=35.7
    pole_longitude=65.5
    lat_0 = 48.8
    delta_lat = 9.6
    lon_0 = 266.1
    delta_lon = 40.1
    map_extent_full=[lon_0-delta_lon, lon_0+delta_lon, lat_0-delta_lat, lat_0+delta_lat]  
    crs_full = ccrs.RotatedPole(pole_latitude=pole_latitude, pole_longitude=pole_longitude)

    #instantiate objects to handle geographical projection of data this needs to be done only once
    # onto geoAxes with this specific crs 
    proj_obj_full = geo_tools.ProjInds(src_lon=longitudes, src_lat=latitudes, 
                                       dest_crs=crs_full, extent=map_extent_full,
                                       image_res=image_res)

    #instantiate figure
    fig = plt.figure(figsize=(fig_w, fig_h))

    proj_obj_list = [proj_obj_full]
    crs_list      = [crs_full]
    extent_list   = [map_extent_full]

    for ii, [this_proj_obj, this_crs, this_extent] in enumerate(zip(proj_obj_list,crs_list,extent_list)):

        #title offset
        xp = .04
        yp = 1.01

        #reference values
        x0 = 0.1/fig_w
        y0 = sp_h + (1.-ii)*(sp_h+rec_h)
        pos = [x0, y0, rec_w, rec_h]
        ax = fig.add_axes(pos, projection=this_crs, extent=this_extent)
        #title
        ax.annotate('reference_val', size=22, xy=(xp, yp), xycoords='axes fraction')
        #print date only once
        #if ii ==0:
            #ax.annotate(this_date.strftime('%Y-%m-%d %Hh%M'), size=24, xy=(xp, 1.2), xycoords='axes fraction')
        #geographical projection of data into axes space
        #ddp
        projected_data = this_proj_obj.project_data(reference_v)
        #draw color figure onto axes
        pr_color_map.plot_data(ax=ax, data=projected_data)
        #format axes and draw geographical boundaries
        plot_geo(ax)
        #palette
        pal_pos = [x0+rec_w+pal_sp, y0, pal_w, rec_h]
        pr_color_map.plot_palette(pal_pos=pal_pos, 
                                  pal_linewidth=0.3, pal_units='[mm/h]',
                                  pal_format='{:3.0f}', equal_legs=True)


        #full grid computation
        if rf is not None:
            x0 = 0.1/fig_w + 1.*(sp_w + rec_w)+ 5.*sp_w
            y0 = sp_h + (1.-ii)*(sp_h+rec_h)
            pos = [x0, y0, rec_w, rec_h]
            ax2 = fig.add_axes(pos, projection=this_crs, extent=this_extent)
            #title
            ax2.annotate('reference fractions', size=22, xy=(xp, yp), xycoords='axes fraction')
            #geographical projection of data into axes space
            #ddp
            projected_data = this_proj_obj.project_data(reference_fractions)
            #draw color figure onto axes
            qi_color_map.plot_data(ax=ax2, data=projected_data)
            #format axes and draw geographical boundaries
            plot_geo(ax2)
            ##palette
            #pal_pos = [x0+rec_w+pal_sp, y0, pal_w, rec_h]
            #qi_color_map.plot_palette(pal_pos=pal_pos, 
            #                             pal_linewidth=0.3, pal_units='[mm]',
            #                             pal_format='{:3.0f}', equal_legs=True)

        #
        if vf is not None:
            x0 = 0.1/fig_w + 1.*(sp_w + rec_w)+ 5.*sp_w
            y0 = sp_h + (0.-ii)*(sp_h+rec_h)
            #
            x0 = 0.1/fig_w + 1.*(sp_w + rec_w)+ 5.*sp_w
            pos = [x0, y0, rec_w, rec_h]
            ax3 = fig.add_axes(pos, projection=this_crs, extent=this_extent)
            #title
            ax3.annotate('verified fractions', size=22, xy=(xp, yp), xycoords='axes fraction')
            #geographical projection of data into axes space
            #ddp
            projected_data = this_proj_obj.project_data(verified_fractions)
            #draw color figure onto axes
            qi_color_map.plot_data(ax=ax3, data=projected_data)
            #format axes and draw geographical boundaries
            plot_geo(ax3)
            ##palette
            #pal_pos = [x0+rec_w+pal_sp, y0, pal_w, rec_h]
            #qi_color_map.plot_palette(pal_pos=pal_pos, 
            #                             pal_linewidth=0.3, pal_units='[mm]',
            #                             pal_format='{:3.0f}', equal_legs=True)


        ##
        ##
        ##plot circles
        #if radius is not None:
        #    delta_pt = 40
        #    subset_lat =  latitudes[::delta_pt,::delta_pt]
        #    subset_lon = longitudes[::delta_pt,::delta_pt]
        #    plot_circles(ax2, subset_lat, subset_lon, radius)

        ##plot circle selection
        #p_file = '/space/hall3/sitestore/eccc/mrd/rpndat/dja001/lmin_investigate/pts.pickle'
        #with open(p_file, 'rb') as f_handle:
        #    rec_dict = pickle.load(f_handle)
        #p_file = '/space/hall3/sitestore/eccc/mrd/rpndat/dja001/lmin_investigate/ll.pickle'
        #with open(p_file, 'rb') as f_handle:
        #    ll_dict = pickle.load(f_handle)

        #proj_cart = ccrs.PlateCarree()
        #print('number of balls', len(rec_dict['pt_list']))
        #for ball_list in rec_dict['pt_list']:
        #    ax2.scatter(ll_dict['longitudes'][ball_list], ll_dict['latitudes'][ball_list], 
        #                transform=proj_cart, color=(203./256.,26./256.,35./256.),zorder=300, s=.05**2.)



        #reference_fractions_int  = np.where(reference_fractions_int < 0., -1., reference_fractions_int)
        #verified_fractions_int   = np.where(verified_fractions_int < 0., -1., verified_fractions_int)

        if rfi is not None:
            x0 = 0.1/fig_w + 2.*(sp_w + rec_w)+ 5.*sp_w
            y0 = sp_h + (1.-ii)*(sp_h+rec_h)
            pos = [x0, y0, rec_w, rec_h]
            ax4 = fig.add_axes(pos, projection=this_crs, extent=this_extent)
            #title
            ax4.annotate('reference interpolated', size=22, xy=(xp, yp), xycoords='axes fraction')
            #geographical projection of data into axes space
            #ddp
            projected_data = this_proj_obj.project_data(reference_fractions_int)
            #draw color figure onto axes
            qi_color_map.plot_data(ax=ax4, data=projected_data)
            #format axes and draw geographical boundaries
            plot_geo(ax4)
            #palette
            pal_pos = [x0+rec_w+pal_sp, y0, pal_w, rec_h]
            qi_color_map.plot_palette(pal_pos=pal_pos, 
                                      pal_linewidth=0.3, pal_units='[unitless]',
                                      pal_format='{:2.1f}')
        #
        if vfi is not None:
            x0 = 0.1/fig_w + 2.*(sp_w + rec_w)+ 5.*sp_w
            y0 = sp_h + (0.-ii)*(sp_h+rec_h)
            #reference fraction
            x0 = 0.1/fig_w + 2.*(sp_w + rec_w)+ 5.*sp_w
            pos = [x0, y0, rec_w, rec_h]
            ax5 = fig.add_axes(pos, projection=this_crs, extent=this_extent)
            #title
            ax5.annotate('verified interpolated', size=22, xy=(xp, yp), xycoords='axes fraction')
            #geographical projection of data into axes space
            #ddp
            projected_data = this_proj_obj.project_data(verified_fractions_int)
            #draw color figure onto axes
            qi_color_map.plot_data(ax=ax5, data=projected_data)
            #format axes and draw geographical boundaries
            plot_geo(ax5)
            #palette
            pal_pos = [x0+rec_w+pal_sp, y0, pal_w, rec_h]
            qi_color_map.plot_palette(pal_pos=pal_pos, 
                                      pal_linewidth=0.3, pal_units='[unitless]',
                                      pal_format='{:2.1f}')

    #make dir if it does not exist
    pic_dir='/space/hall3/sitestore/eccc/mrd/rpndat/dja001/test_new_griddedobs/' 
    if not os.path.isdir(pic_dir):
        domutils._py_tools.parallel_mkdir(pic_dir)

    #figure name
    #if not np.isclose(leadtime.seconds, 0):
    #    lt_minutes = leadtime.days*1440. + np.floor(leadtime.seconds/60.)
    #    svg_name = pic_dir+'/compare_fields_'+validity_date.strftime('%Y%m%d%H%M')+'{:+06.0f}m_th{:04.1f}'.format(lt_minutes, threshold)+'.svg'
    #else:
    #    svg_name = pic_dir+'/compare_fields_'+validity_date.strftime('%Y%m%d%H%M')+'_th{:03.0f}'.format(threshold)+'.svg'
    svg_name = pic_dir+'/compare_fractions_radius'+'{:05.2f}'.format(radius)+'.svg'
    plt.savefig(svg_name, dpi=image_dpi)
    plt.close(fig)
    py_tools.lmroman(svg_name)
    #py_tools.convert(svg_name,'gif', del_orig=True, density=400, geometry='50%')
    print('done with: ', svg_name)


if __name__ == "__main__":     
    main()




def twod_panels(params, domain, twod_avg_list, leadtime_minutes,
                hist_bin_edges, avg_reference_hists, avg_verified_hists):
    """plot 2D pannels statistics

    what is being plotted is defined in params.twod_panels from the bottom up
    """

    import numpy as np
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import matplotlib.ticker
    import domcmc.fst_tools as fst_tools
    import domutils.legs as legs
    import domutils.geo_tools as geo_tools
    import domutils.radar_tools as radar_tools
    import domutils._py_tools as py_tools
    import gridded_obs.plot 

    n_stat_types  = len(params.twod_panels)
    n_panels = len(twod_avg_list)

    #missing val
    missing = -9999.

    #figure dimensions
    ratio  = 0.75
    rec_w  = 4.           # Horizontal size of a panel
    rec_h  = ratio*rec_w  # Vertical size of a panel
    sp_w   = 1.2           # horizontal space between panels
    sp_h   = 1.           # vertical space between panels
    pal_sp = .05          # space between panel and palette
    pal_w  = .17          # width of palette
    tit_h  = 1.6 + 0.5*np.ceil(len(params.exp_list)/2.)  # height of title
    #size of figure
    fig_w  = sp_w + n_panels*(rec_w + sp_w) +  sp_w
    fig_h  = sp_h + n_stat_types*(rec_h+sp_h) + tit_h + .9
    #normalize sizes relative to figure
    rec_w  = rec_w / fig_w
    rec_h  = rec_h / fig_h
    sp_w   = sp_w / fig_w
    sp_h   = sp_h / fig_h
    pal_sp = pal_sp / fig_w
    pal_w  = pal_w / fig_w
    tit_h  = tit_h / fig_h
    #pixel resolution of image being generated
    grid_w_pts = 1200.
    image_dpi = grid_w_pts/(rec_w*fig_w)
    image_res = [grid_w_pts,ratio*grid_w_pts]

    #larger typeface
    mpl.rcParams.update({'font.size': 20})
    # Use this for editable text in svg
    mpl.rcParams['text.usetex']  = False
    mpl.rcParams['svg.fonttype'] = 'none'
    # Hi def figure
    mpl.rcParams['figure.dpi'] = 600

    #instantiate figure
    fig = plt.figure(figsize=(fig_w, fig_h))

    #header, legend and other explanations
    gridded_obs.plot.plot_header(fig, domain, params,
                                 fig_w  , fig_h  ,
                                 rec_w  , rec_h,
                                 sp_w   , sp_h,
                                 pal_sp , pal_w,
                                 tit_h  )

    for nn, stat_type in enumerate(params.twod_panels):
        y0 = sp_h + nn*(rec_h + sp_h) 
        
        params.ylim_hist = None
        if params.ylim_hist is not None:
            y_lim = params.ylim_hist
        else:
            y_lim = gridded_obs.plot.autorange(avg_reference_hists)
            y_lim = (0., y_lim[1])   #no bottom headspace for histograms

        for ii, lt_inds in enumerate(twod_avg_list):
            x0 = sp_w + ii*(rec_w + sp_w) 
            pos = [x0, y0, rec_w, rec_h]


            #Setup the general look 
            #Ticks length and width
            major_height = 5
            major_width = .3
            minor_height = 2
            minor_width = .3
            #axes containing data
            ax = fig.add_axes(pos, label = 'twod_plot', zorder=10)
            ax.set_facecolor((0.,0.,0.,0.))
            #ax.set_xlim((leadtime_hours[0], leadtime_hours[-1]))
            ax.set_ylim(y_lim)
            #ax.tick_params(axis='both', which='both', direction='in', top='on', right=True)
            #ax.set_xticks(xticks)
            #ax.set_yticks(this_var_info['ticks']) 
            #ax.tick_params('both', length=major_height, width=major_width, which='major')
            #ax.tick_params('both', length=minor_height, width=minor_width, which='minor')
            #if nn == 0:
            #    ax.set_xlabel('Leadtime (h)')
            #else:
            #    ax.set_xticklabels([])
            #ax.set_ylabel(unit_dict[score])
            ax.set_xscale('log')
            ax.set_xlim((1e-2,1e2))
            ## set x ticks
            x_major = matplotlib.ticker.FixedLocator([1e-2, 1e-1, 1., 1e1, 1e2])
            #major = matplotlib.ticker.LogLocator(base = 10.0, numticks = 5)
            ax.xaxis.set_major_locator(x_major)
            x_minor = matplotlib.ticker.LogLocator(base = 10.0, subs = np.arange(1.0, 10.0) * 0.1, numticks = 10)
            #x_minor = matplotlib.ticker.Formatter
            ax.xaxis.set_minor_locator(x_minor)
            #ax.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%f2.0'))
            ax.set_xticklabels(['0.01', '0.1', '1', '10', '100' ])
            ax.xaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())

            #ax info
            minute_0 = leadtime_minutes[lt_inds[0]]
            minute_f = leadtime_minutes[lt_inds[-1]]
            bound_str = '[{:4d}m,{:4d}m]'.format(minute_0, minute_f)
            ax.annotate(bound_str, (.50, .96), xycoords='axes fraction', ha='left', va='top', fontsize=18)

            left  = hist_bin_edges[:-1]
            right = hist_bin_edges[1:]
            xx = np.array([left,right]).T.flatten()

            #for reference_data
            bin_v = avg_reference_hists[ii,:]
            yy = np.array([bin_v,bin_v]).T.flatten()
            ax.plot(  xx, yy, color='black')

            for kk, experiment in enumerate(params.exp_list):
                bin_v = avg_verified_hists[kk,ii,:]
                yy = np.array([bin_v,bin_v]).T.flatten()
                ax.plot( xx, yy,color=params.exp_color[kk], 
                                linestyle=params.exp_linestyle[kk], 
                                linewidth=params.exp_linewidth[kk], )
    #save image
    date_dir = params.score_dir+ params.date_list[0].strftime('%Y%m%d%H')+'_to_'+ params.date_list[-1].strftime('%Y%m%d%H')
    py_tools.parallel_mkdir(date_dir)
    svg_name = date_dir+'/twod_panels__'+domain+'.svg'
    plt.savefig(svg_name, dpi=image_dpi)
    plt.close(fig)
    py_tools.lmroman(svg_name)
    #py_tools.convert(svg_name,'gif', del_orig=True, density=400, geometry='50%')
    print('done with: ', svg_name)


